#! /usr/env/python

"""
This module attempts to "component-ify" GT's Fastscape stream power erosion.
Created DEJH, March 2014.
"""

import numpy
from landlab import ModelParameterDictionary
from landlab.core.model_parameter_dictionary import MissingKeyError, ParameterValueError
from landlab.field.scalar_data_fields import FieldError
from scipy import weave
from scipy.optimize import newton, fsolve
#from scipy.weave.build_tools import CompileError
UNDEFINED_INDEX = numpy.iinfo(numpy.int32).max

class SPEroder(object):
    '''
    This class uses the Braun-Willett Fastscape approach to calculate the amount
    of erosion at each node in a grid, following a stream power framework.
    
    On initialization, it takes *grid*, a reference to a ModelGrid, and
    *input_stream*, a string giving the filename (and optionally, path) of the
    required input file.
    
    It needs to be supplied with the key variables:
    
        *K_sp*
        
        *m_sp*
    
    ...which it will draw from the supplied input file. *n_sp*  can be any 
    value ~ 0.5<n_sp<4., but note that performance will be EXTREMELY degraded
    if n<1.
    
    If you want to supply a spatial variation in K, set K_sp to the string
    'array', and pass a field name or array to the erode method's K_if_used
    argument.
    
    *dt*, *rainfall_intensity*, and *value_field* are optional variables.
    
    *dt* is a fixed timestep, and *rainfall_intensity* is a parameter which 
    modulates K_sp (by a product, r_i**m_sp) to reflect the direct influence of
    rainfall intensity on erosivity. *value_field* is a string giving the name
    of the field containing the elevation data in the grid. It defaults to
    'topographic_elevation' if not supplied.
    
    This module assumes you have already run 
    :func:`landlab.components.flow_routing.route_flow_dn.FlowRouter.route_flow`
    in the same timestep. It looks for 'upstream_ID_order', 
    'links_to_flow_receiver', 'drainage_area', 'flow_receiver', and
    'topographic_elevation' at the nodes in the grid. 'drainage_area' should
    be in area upstream, not volume (i.e., set runoff_rate=1.0 when calling
    FlowRouter.route_flow).
    
    If dt is not supplied, you must call gear_timestep(dt_in, rain_intensity_in)
    each iteration to set these variables on-the-fly (rainfall_intensity will be
    overridden if supplied in the input file).
    If dt is supplied but rainfall_intensity is not, the module will assume you
    mean r_i = 1.
    
    The primary method of this class is :func:`erode`.
    '''
    
    def __init__(self, grid, input_stream):
        self.grid = grid
        inputs = ModelParameterDictionary(input_stream)
        
        #User sets:
        try:
            self.K = inputs.read_float('K_sp')
        except ParameterValueError:
            self.use_K = True
        else:
            self.use_K = False
            
        self.m = inputs.read_float('m_sp')
        try:
            self.n = inputs.read_float('n_sp')
        except:
            self.n = 1.
        try:
            self.dt = inputs.read_float('dt')
        except: #if dt isn't supplied, it must be set by another module, so look in the grid
            print 'Set dynamic timestep from the grid. You must call gear_timestep() to set dt each iteration.'
        try:
            self.r_i = inputs.read_float('rainfall_intensity')
        except:
            self.r_i = 1.
        try:
            self.value_field = inputs.read_str('value_field')
        except:
            self.value_field = 'topographic_elevation'
            
        #make storage variables
        self.A_to_the_m = grid.create_node_array_zeros()
        self.alpha = grid.empty(centering='node')
        self.alpha_by_flow_link_lengthtothenless1 = numpy.empty_like(self.alpha)
        
        self.grid.node_diagonal_links() #calculates the number of diagonal links
        
        if self.n != 1.:
            #raise ValueError('The Braun Willett stream power algorithm requires n==1. at the moment, sorry...')
            self.nonlinear_flag = True
            if self.n<1.:
                print "***WARNING: With n<1 performance of the Fastscape algorithm is slow!***"
        else:
            self.nonlinear_flag = False
        
        self.weave_flag = grid.weave_flag
        
        def func_for_newton(x, last_step_elev, receiver_elev, alpha_by_flow_link_lengthtothenless1, n):
            y = x - last_step_elev + alpha_by_flow_link_lengthtothenless1*(x-receiver_elev)**n
            return y
        
        def func_for_newton_diff(x, last_step_elev, receiver_elev, alpha_by_flow_link_lengthtothenless1, n):
            y = 1. + n*alpha_by_flow_link_lengthtothenless1*(x-receiver_elev)**(n-1.)
            return y
        
        self.func_for_newton = func_for_newton
        self.func_for_newton_diff = func_for_newton_diff

    def gear_timestep(self, dt_in, rainfall_intensity_in=None):
        self.dt = dt_in
        if rainfall_intensity_in is not None:
            self.r_i = rainfall_intensity_in
        return self.dt, self.r_i
        

    def erode(self, grid_in, K_if_used=None):
        """
        This method implements the stream power erosion, following the Braun-
        Willett (2013) implicit Fastscape algorithm. This should allow it to
        be stable against larger timesteps than an explicit stream power scheme.
        
        The method takes *grid*, a reference to the model grid.
        Set 'K_if_used' as a field name or nnodes-long array if you set K_sp as
        'array' during initialization.
        
        It returns the grid, in which it will have modified the value of 
        *value_field*, as specified in component initialization.
        """
        
        #self.grid = grid_in #the variables must be stored internally to the grid, in fields
        upstream_order_IDs = self.grid['node']['upstream_ID_order']
        #ordered_receivers = self.grid['node']['flow_receiver'][upstream_order_IDs]  #"j" in GT's sketch
        #nonboundaries = numpy.not_equal(upstream_order_IDs, ordered_receivers)
        z = self.grid['node'][self.value_field]
        #interior_nodes = numpy.greater_equal(self.grid['node']['links_to_flow_receiver'], -1)
        #interior_nodes = (self.grid['node']['links_to_flow_receiver'][upstream_order_IDs])[nonboundaries]
        #flow_link_lengths = self.grid.link_length[interior_nodes]
        ##defined_flow_receivers = numpy.greater_equal(self.grid['node']['links_to_flow_receiver'],-1)
        defined_flow_receivers = numpy.not_equal(self.grid['node']['links_to_flow_receiver'],UNDEFINED_INDEX)
        #flow_link_lengths = numpy.zeros_like(self.alpha)
        flow_link_lengths = self.grid.link_length[self.grid['node']['links_to_flow_receiver'][defined_flow_receivers]]
        
        if K_if_used!=None:
            assert self.use_K, "An array of erodabilities was provided, but you didn't set K_sp to 'array' in your input file! Aborting..."
            try:
                self.K = self.grid.at_node[K_if_used][defined_flow_receivers]
            except TypeError:
                self.K = K_if_used[defined_flow_receivers]
        
        #regular_links = numpy.less(self.grid['node']['links_to_flow_receiver'][defined_flow_receivers],self.grid.number_of_links)
        #flow_link_lengths[defined_flow_receivers][regular_links] = self.grid.link_length[(self.grid['node']['links_to_flow_receiver'])[defined_flow_receivers][regular_links]]
        #diagonal_links = numpy.logical_not(regular_links)
        #flow_link_lengths[defined_flow_receivers][diagonal_links] = numpy.sqrt(self.grid.node_spacing*self.grid.node_spacing)
        numpy.power(self.grid['node']['drainage_area'], self.m, out=self.A_to_the_m)
        #self.alpha[nonboundaries] = self.K * self.dt * self.A_to_the_m[nonboundaries] / flow_link_lengths
        self.alpha[defined_flow_receivers] = self.r_i**self.m * self.K * self.dt * self.A_to_the_m[defined_flow_receivers] / flow_link_lengths

        flow_receivers = self.grid['node']['flow_receiver']
        n_nodes = upstream_order_IDs.size
        alpha = self.alpha
        if self.nonlinear_flag==False: #n==1
            if self.weave_flag:
                code = """
                    int current_node;
                    int j;
                    for (int i = 0; i < n_nodes; i++) {
                        current_node = upstream_order_IDs[i];
                        j = flow_receivers[current_node];
                        if (current_node != j) {
                            z[current_node] = (z[current_node] + alpha[current_node]*z[j])/(1.0+alpha[current_node]);
                        }
                    }
                """
                weave.inline(code, ['n_nodes', 'upstream_order_IDs', 'flow_receivers', 'z', 'alpha'])
            else:
                for i in upstream_order_IDs:
                    j = flow_receivers[i]
                    if i != j:
                        z[i] = (z[i] + alpha[i]*z[j])/(1.0+alpha[i])
        else: #general, nonlinear n case
            self.alpha_by_flow_link_lengthtothenless1[defined_flow_receivers] = alpha[defined_flow_receivers]/flow_link_lengths**(self.n-1.)
            alpha_by_flow_link_lengthtothenless1 = self.alpha_by_flow_link_lengthtothenless1
            n = float(self.n)
            if self.weave_flag:
                if n<1.:
                    #this is SLOOOOOOOOOOOW...
                    for i in upstream_order_IDs:
                        j = flow_receivers[i]
                        func_for_newton = self.func_for_newton
                        func_for_newton_diff = self.func_for_newton_diff
                        if i != j:
                            z[i] = fsolve(func_for_newton, z[i], args=(z[i], z[j], alpha_by_flow_link_lengthtothenless1[i], n))
                else:
                    code = """
                        int current_node;
                        int j;
                        double current_z;
                        double previous_z;
                        double elev_diff;
                        double elev_diff_tothenless1;
                        for (int i = 0; i < n_nodes; i++) {
                            current_node = upstream_order_IDs[i];
                            j = flow_receivers[current_node];
                            previous_z = z[current_node];
                            if (current_node != j) {
                                while (1) {
                                    elev_diff = previous_z-z[j];
                                    elev_diff_tothenless1 = pow(elev_diff, n-1.); //this isn't defined if in some iterations the elev_diff goes -ve
                                    current_z = previous_z - (previous_z - z[current_node] + alpha_by_flow_link_lengthtothenless1[current_node]*elev_diff_tothenless1*elev_diff)/(1.+n*alpha_by_flow_link_lengthtothenless1[current_node]*elev_diff_tothenless1);
                                    if (abs((current_z - previous_z)/current_z) < 1.48e-08) break;
                                    previous_z = current_z;
                                }
                                z[current_node] = current_z;
                            }
                        }
                    """
                    weave.inline(code, ['n_nodes', 'upstream_order_IDs', 'flow_receivers', 'z', 'alpha_by_flow_link_lengthtothenless1', 'n'], headers=["<math.h>"])
            else:
                for i in upstream_order_IDs:
                    j = flow_receivers[i]
                    func_for_newton = self.func_for_newton
                    func_for_newton_diff = self.func_for_newton_diff
                    if i != j:
                        if n>=1.:
                            z[i] = newton(func_for_newton, z[i], fprime=func_for_newton_diff, args=(z[i], z[j], alpha_by_flow_link_lengthtothenless1[i], n), maxiter=10)
                        else:
                            z[i] = fsolve(func_for_newton, z[i], args=(z[i], z[j], alpha_by_flow_link_lengthtothenless1[i], n))
        #self.grid['node'][self.value_field] = z
        
        return self.grid




#! /usr/bin/env python
import numpy as np
import itertools

from ..grid.base import (CORE_NODE, FIXED_GRADIENT_BOUNDARY,
                         FIXED_VALUE_BOUNDARY, TRACKS_CELL_BOUNDARY,
                         CLOSED_BOUNDARY, BAD_INDEX_VALUE)


def node_count(shape):
    """
    The total number of nodes in a structured grid with dimensions given
    by the tuple, *shape*. Where *shape* is the number of node rows and
    node columns.

    >>> from landlab.utils.structured_grid import node_count
    >>> node_count((3, 4))
    12
    """
    assert(len(shape) == 2)
    return shape[0] * shape[1]


def interior_node_count(shape):
    """
    Return the count of the number of interior nodes of a structured grid
    of dimensions, *shape*.

    >>> from landlab.utils.structured_grid import (node_count,
    ...                                            interior_node_count)
    >>> node_count((2, 4))
    8
    >>> interior_node_count((2, 4))
    0
    >>> interior_node_count((1, 4))
    0
    >>> interior_node_count((3, 4))
    2
    """
    assert(len(shape) == 2)
    try:
        assert(np.min(shape) > 2)
    except AssertionError:
        return 0
    else:
        return (shape[0] - 2) * (shape[1] - 2)


def cell_count(shape):
    """
    The total number of cells in a structured grid with dimensions, *shape*.
    Where *shape* is a tuple that gives the dimensions of the grid as number
    of rows of nodes followed by number of columns of nodes.

    >>> from landlab.utils.structured_grid import cell_count
    >>> cell_count((3, 4))
    2
    >>> cell_count((1, 4))
    0
    """
    assert(len(shape) == 2)

    if np.min(shape) > 2:
        return (shape[0] - 2) * (shape[1] - 2)
    else:
        return 0


def active_cell_count(shape):
    """
    Number of active cells. By default, all cells are active so this is
    the same as cell_count. (active = core+open boundary)
    """
    return cell_count(shape)

def core_cell_count(shape):
    """
    Number of core cells. By default, all cells are core so this is
    the same as cell_count.
    """
    return cell_count(shape)

def active_link_count(shape):
    """
    Number of active links in a structured grid with dimensions, *shape*.
    A link is active if it connects to at least one active node.

    >>> from landlab.utils.structured_grid import link_count, active_link_count
    >>> link_count((3, 2))
    7
    >>> active_link_count((3, 2))
    0
    >>> active_link_count((3, 4))
    7
    """
    assert(len(shape) == 2)
    try:
        assert(np.min(shape) > 2)
    except AssertionError:
        return 0
    else:
        return 2 * shape[0] * shape[1] - 3 * (shape[0] + shape[1]) + 4


def link_count(shape):
    """
    Total (active and inactive) number of links in a structured grid with
    dimensions, *shape*. This is the number of to-links and from-links, not
    the total of the two.

    >>> from landlab.utils.structured_grid import link_count
    >>> link_count((3,2))
    7
    """
    assert(len(shape) == 2)
    return shape[1] * (shape[0] - 1) + shape[0] * (shape[1] - 1)


def vertical_link_count(shape):
    assert(len(shape) == 2)
    return (shape[0] - 1) * shape[1]


def horizontal_link_count(shape):
    assert(len(shape) == 2)
    return shape[0] * (shape[1] - 1)


def boundary_cell_count(shape):
    """
    .. deprecated:: 0.6
    Deprecated due to imprecise terminology; this function makes little sense.
    Use perimeter_node_count() instead.
    Number of cells that are on the perimeter of a structured grid with
    dimensions, *shape*, and thus boundary cells. In fact, cells centered on
    boundary nodes are not really cells. If they were, though, this is how
    many there would be.

    **** Shouldn't be deprecated. This routine returns the cells on the
    boundary. Not the cells surrounding boundary nodes because there aren't
    cells around boundary nodes by definition as previously understood.
      - SN  30Nov14  ****

    Examples
    --------
    >>> from landlab.utils.structured_grid import boundary_cell_count
    >>> boundary_cell_count((3, 4))
    10
    """
    assert(len(shape) == 2)
    return 2 * (shape[0] - 2) + 2 * (shape[1] - 2) + 4


def perimeter_node_count(shape):
    """
    Number of nodes that are on the perimeter of a structured grid with
    dimensions, *shape*, and thus boundary nodes.

    >>> from landlab.utils.structured_grid import perimeter_node_count
    >>> perimeter_node_count((3, 4))
    10
    """
    assert(len(shape) == 2)
    return 2 * (shape[0] - 2) + 2 * (shape[1] - 2) + 4


def interior_cell_count(shape):
    """
    Number of interior cells. Since cells are only defined on interior nodes,
    this is the same as cell_count.
    """
    return cell_count(shape)


def face_count(shape):
    """
    Total number of faces in a structured grid with dimensions, *shape*. Each
    cell has four faces, and shared faces only count once.

    >>> from landlab.utils.structured_grid import face_count
    >>> face_count((3, 4))
    7
    """
    assert(len(shape) == 2)
    if np.min(shape) > 2:
        return ((shape[0] - 1) * (shape[1] - 2) +
                (shape[0] - 2) * (shape[1] - 1))
    else:
        return 0


def active_face_count(shape):
    """
    Total number of active faces in a structured grid with dimensions,
    *shape*. Each cell has four faces, and shared faces only count once.
    An active face is one that has a corresponing active link.

    >>> from landlab.utils.structured_grid import active_face_count
    >>> active_face_count((3, 4))
    7
    """
    return face_count(shape)


def top_index_iter(shape):
    """
    Iterator for the top boundary indices of a structured grid.
    """
    return xrange(shape[1] * (shape[0] - 1), shape[0] * shape[1])


def bottom_index_iter(shape):
    """
    Iterator for the bottom boundary indices of a structured grid.
    """
    return xrange(0, shape[1])


def left_index_iter(shape):
    """
    Iterator for the left boundary indices of a structured grid.
    """
    return xrange(0, shape[0] * shape[1], shape[1])


def right_index_iter(shape):
    """
    Iterator for the right boundary indices of a structured grid.
    """
    return xrange(shape[1] - 1, shape[0] * shape[1], shape[1])


def left_right_iter(shape, *args):
    """
    Iterator for the left and right boundary indices of a structured grid.
    This iterates over the indices in order rather than iterating all of
    the left boundary and then all of the right boundary.

    >>> import numpy as np
    >>> from landlab.utils.structured_grid import left_right_iter
    >>> np.fromiter(left_right_iter((4, 3)), dtype=np.int)
    array([ 0,  2,  3,  5,  6,  8,  9, 11])
    >>> np.fromiter(left_right_iter((4, 3), 2), dtype=np.int)
    array([0, 2, 3, 5])
    >>> np.fromiter(left_right_iter((4, 3), 2, 4), dtype=np.int)
    array([ 6,  8,  9, 11])
    >>> np.fromiter(left_right_iter((4, 3), 1, 4, 2), dtype=np.int)
    array([ 3,  5,  9, 11])
    """
    if len(args) == 0:
        iter_rows = xrange(0, shape[0], 1)
    elif len(args) == 1:
        iter_rows = xrange(0, args[0], 1)
    elif len(args) == 2:
        iter_rows = xrange(args[0], args[1], 1)
    elif len(args) == 3:
        iter_rows = xrange(args[0], args[1], args[2])

    for row in iter_rows:
        yield row * shape[1]
        yield row * shape[1] + shape[1] - 1


def bottom_top_iter(shape):
    """
    Iterates of the bottom indices and then the top indices of a structured
    grid.

    >>> import numpy as np
    >>> from landlab.utils.structured_grid import bottom_top_iter
    >>> np.fromiter(bottom_top_iter((4, 3)), dtype=np.int)
    array([ 0,  1,  2,  9, 10, 11])
    """
    return itertools.chain(bottom_index_iter(shape),
                           top_index_iter(shape))


def boundary_iter(shape):
    """
    .. deprecated:: 0.6
    Deprecated due to imprecise terminology. This is really perimeter_iter
    (see below).
    Iterates over all of the boundary node indices of a structured grid in
    order.

    >>> import numpy as np
    >>> from landlab.utils.structured_grid import boundary_iter
    >>> np.fromiter(boundary_iter((4, 3)), dtype=np.int)
    array([ 0,  1,  2,  3,  5,  6,  8,  9, 10, 11])
    """
    return itertools.chain(bottom_index_iter(shape),
                           left_right_iter(shape, 1, shape[0] - 1),
                           top_index_iter(shape))

def perimeter_iter(shape):
    """
    Iterates over all of the perimeter node indices of a structured grid in
    order.

    >>> import numpy as np
    >>> from landlab.utils.structured_grid import perimeter_iter
    >>> np.fromiter(perimeter_iter((4, 3)), dtype=np.int)
    array([ 0,  1,  2,  3,  5,  6,  8,  9, 10, 11])
    """
    return itertools.chain(bottom_index_iter(shape),
                           left_right_iter(shape, 1, shape[0] - 1),
                           top_index_iter(shape))

def boundary_nodes(shape):
    """
    .. deprecated:: 0.6
    Deprecated due to imprecise terminology. This is really perimeter_iter
    (see below).
    An array of the indices of the boundary nodes.

    >>> from landlab.utils.structured_grid import boundary_nodes
    >>> boundary_nodes((3, 4))
    array([ 0,  1,  2,  3,  4,  7,  8,  9, 10, 11])
    """
    return np.fromiter(boundary_iter(shape), dtype=np.int)

def perimeter_nodes(shape):
    """
    An array of the indices of the perimeter nodes of a structured grid.

    >>> from landlab.utils.structured_grid import perimeter_nodes
    >>> perimeter_nodes((3, 4))
    array([ 0,  1,  2,  3,  4,  7,  8,  9, 10, 11])
    """
    return np.fromiter(perimeter_iter(shape), dtype=np.int)

def corners(shape):
    """
    An array of the indices of the grid corner nodes.
    """
    return np.array([0, shape[1]-1, shape[1]*(shape[0]-1), shape[1]*shape[0]-1])

def bottom_edge_node_ids(shape):
    return np.fromiter(bottom_index_iter(shape), dtype=np.int)


def top_edge_node_ids(shape):
    return np.fromiter(top_index_iter(shape), dtype=np.int)


def left_edge_node_ids(shape):
    return np.fromiter(left_index_iter(shape), dtype=np.int)


def right_edge_node_ids(shape):
    return np.fromiter(right_index_iter(shape), dtype=np.int)


def interior_iter(shape):
    """
    Iterate over the interior nodes of a structured grid.

    >>> import numpy as np
    >>> from landlab.utils.structured_grid import interior_iter
    >>> np.fromiter(interior_iter((4, 3)), dtype=np.int)
    array([4, 7])
    """
    interiors = []
    interiors_per_row = shape[1] - 2
    for row in xrange(shape[1] + 1, shape[1] * (shape[0] - 1), shape[1]):
        interiors.append(xrange(row , row + interiors_per_row))
    return itertools.chain(*interiors)


def interior_nodes(shape):
    return np.fromiter(interior_iter(shape), dtype=np.int)


def node_coords(shape, *args):
    """
    Get x, y coordinates for nodes in a structured grid with dimensions,
    *shape*. Use the optional argument *spacing* to give the spacing in each
    dimension, and *origin* the start of the coordinates in each dimension.

    >>> from landlab.utils.structured_grid import node_coords
    >>> (cols, rows) = node_coords((3, 2))
    >>> rows
    array([ 0.,  0.,  1.,  1.,  2.,  2.])
    >>> cols
    array([ 0.,  1.,  0.,  1.,  0.,  1.])
    """
    try:
        spacing = args[0]
    except IndexError:
        spacing = np.ones(len(shape), dtype=np.float)
    else:
        assert(len(spacing) == len(shape))

    try:
        origin = args[1]
    except IndexError:
        origin = np.zeros(len(shape), dtype=np.float)
    else:
        assert(len(origin) == len(origin))

    node_count = np.prod(shape)

    row_y = np.arange(shape[0]) * spacing[0] + origin[0]
    col_x = np.arange(shape[1]) * spacing[1] + origin[1]

    (node_x, node_y) = np.meshgrid(col_x, row_y)

    node_x.shape = (node_count, )
    node_y.shape = (node_count, )

    return (node_x, node_y)


def active_cell_index(shape):
    """
    For many instances, core_cell_index() may be preferred.
    Ordered indices of the active (core+open boundary) cells of a structured
    grid.
    """
    return np.arange(active_cell_count(shape))


def core_cell_index(shape):
    """
    Ordered indices of the core cells of a structured grid.
    """
    return np.arange(core_cell_count(shape))


def active_cell_node(shape):
    """
    For many instances, core_cell_node() may be preferred.
    Indices of the nodes belonging to each active (core + open boundary) cell.
    Since all cells are active in the default case, this is the same as
    node_index_at_cells.

    >>> from landlab.utils.structured_grid import active_cell_node
    >>> active_cell_node((4,3))
    array([4, 7])
    """
    return node_index_at_cells(shape)


def core_cell_node(shape):
    """
    Indices of the nodes belonging to each core cell.
    Since all cells are core in the default case, this is the same as
    node_index_at_cells.

    >>> from landlab.utils.structured_grid import core_cell_node
    >>> core_cell_node((4,3))
    array([4, 7])
    """
    return node_index_at_cells(shape)


def active_cell_index_at_nodes(shape, boundary_node_index=BAD_INDEX_VALUE):
    """
    .. deprecated:: 0.6
    Up to date, unambiguous terminology core_cell_index_at_nodes() preferred.
    Indices of the active cells associated with the nodes of the structured grid.
    For nodes that don't have a cell (that is, boundary nodes) set indices
    to BAD_INDEX_VALUE. Use the *boundary_node_index* keyword to change
    the value of indices to boundary nodes.

    Note that all three functions [X_]cell_index_at_nodes are equivalent.

    >>> from landlab.utils.structured_grid import active_cell_index_at_nodes
    >>> active_cell_index_at_nodes((3, 4), boundary_node_index=-1) # doctest: +NORMALIZE_WHITESPACE
    array([-1, -1, -1, -1,
           -1,  0,  1, -1,
           -1, -1, -1, -1])
    """
    n_nodes = node_count(shape)

    node_ids = np.empty(n_nodes, dtype=np.int)

    node_ids[boundary_nodes(shape)] = boundary_node_index
    node_ids[interior_nodes(shape)] = np.arange(interior_node_count(shape))

    return node_ids

def core_cell_index_at_nodes(shape, boundary_node_index=BAD_INDEX_VALUE):
    """
    Indices of the core cells associated with the nodes of the structured grid.
    For nodes that don't have a cell (that is, boundary nodes) set indices
    to BAD_INDEX_VALUE. Use the *boundary_node_index* keyword to change
    the value of indices to boundary nodes.

    Note that all three functions [X_]cell_index_at_nodes are equivalent.

    >>> from landlab.utils.structured_grid import core_cell_index_at_nodes
    >>> core_cell_index_at_nodes((3, 4), boundary_node_index=-1) # doctest: +NORMALIZE_WHITESPACE
    array([-1, -1, -1, -1,
           -1,  0,  1, -1,
           -1, -1, -1, -1])
    """
    n_nodes = node_count(shape)

    node_ids = np.empty(n_nodes, dtype=np.int)

    node_ids[perimeter_nodes(shape)] = boundary_node_index
    node_ids[interior_nodes(shape)] = np.arange(interior_node_count(shape))

    return node_ids

def cell_index_at_nodes(shape, boundary_node_index=BAD_INDEX_VALUE):
    """
    Indices of the core cells associated with the nodes of the structured grid.
    For nodes that don't have a cell (that is, boundary nodes) set indices
    to BAD_INDEX_VALUE. Use the *boundary_node_index* keyword to change
    the value of indices to boundary nodes.

    Note that all three functions [X_]cell_index_at_nodes are equivalent.

    >>> from landlab.utils.structured_grid import cell_index_at_nodes
    >>> cell_index_at_nodes((3, 4), boundary_node_index=-1) # doctest: +NORMALIZE_WHITESPACE
    array([-1, -1, -1, -1,
           -1,  0,  1, -1,
           -1, -1, -1, -1])
    """
    n_nodes = node_count(shape)

    node_ids = np.empty(n_nodes, dtype=np.int)

    node_ids[perimeter_nodes(shape)] = boundary_node_index
    node_ids[interior_nodes(shape)] = np.arange(interior_node_count(shape))

    return node_ids


def node_index_at_cells(shape):
    """
    Indices of the nodes belonging to each cell.

    >>> from landlab.utils.structured_grid import node_index_at_cells
    >>> node_index_at_cells((4, 3))
    array([4, 7])
    """
    node_ids = np.arange(node_count(shape))
    node_ids.shape = shape

    cell_node = node_ids[1:-1, 1:-1].copy()
    cell_node.shape = ((shape[0] - 2) * (shape[1] - 2), )

    return cell_node


def node_index_at_link_ends(shape):
    node_ids = np.arange(np.prod(shape))
    node_ids.shape = shape

    return (node_index_at_link_head(node_ids),
            node_index_at_link_tail(node_ids))


def inlink_index_at_node(shape):
    return inlinks(shape, return_count=False)


def outlink_index_at_node(shape):
    return outlinks(shape, return_count=False)


def node_index_at_link_tail(node_ids):
    vertical_links = node_ids[1:, :]
    horizontal_links = node_ids[:, 1:]
    return np.concatenate((vertical_links.flat, horizontal_links.flat))


def node_index_at_link_head(node_ids):
    vertical_links = node_ids[:-1, :]
    horizontal_links = node_ids[:, :-1]
    return np.concatenate((vertical_links.flat, horizontal_links.flat))


def face_index_at_links(shape, actives=None,
                        inactive_link_index=BAD_INDEX_VALUE):
    """
    Returns an array that maps link ids to face ids. For inactive links,
    which do not have associated faces, set their ids to
    *inactive_link_index*. Use the *actives* keyword to specify an array that
    contains the ids of all active links in the grid. The default assumes
    that only the perimeter nodes are inactive.


    >>> from landlab.utils.structured_grid import face_index_at_links
    >>> faces = face_index_at_links((3, 4), inactive_link_index=-1)
    >>> faces # doctest: +NORMALIZE_WHITESPACE
    array([-1,  0,  1, -1, -1,  2,  3,
           -1, -1, -1, -1,  4,  5,  6, -1, -1, -1])

    """
    if actives is None:
        actives = active_links(shape)

    num_links = link_count(shape)

    link_faces = np.empty(num_links, dtype=np.int)
    link_faces.fill(inactive_link_index)
    link_faces[actives] = np.arange(len(actives))

    return link_faces


def node_status(shape, boundary_status=FIXED_VALUE_BOUNDARY):
    """
    The statuses of the nodes in a structured grid with dimensions, *shape*.
    Use the *boundary_status* keyword to specify the status of the top,
    bottom, left and right boundary nodes.
    """
    status = np.empty(np.prod(shape), dtype=np.int8)

    status[interior_nodes(shape)] = CORE_NODE
    status[boundary_nodes(shape)] = boundary_status

    return status


def active_links(shape, node_status_array=None, link_nodes=None):
    """Link IDs for active links of a structured quad grid.

    Return the link IDs for links that are *active* in a structured grid of quadrilaterals. Use the
    *node_status_array* keyword to specify the status for each of the grid's nodes. If not given, each of the
    perimeter nodes is assumed to be `FIXED_VALUE_BOUNDARY`.

    Use the *link_nodes* keyword to provide, as a tuple of arrays, that give the *from-node* and the *to-node*
    for each for each link in the grid.

    Parameters
    ----------
    shape : tuple
        Shape of grid as number of node rows and columns.
    node_status_array : array_like, optional
        Status of each grid node.
    link_nodes : array_like, optional

    Examples
    --------
    Because, by default, the perimeter nodes are `FIXED_VALUE_BOUNDARY` nodes, only links attached to the
    interior nodes are *active*.

    >>> from landlab.utils.structured_grid import active_links
    >>> from landlab import CLOSED_BOUNDARY, CORE_NODE
    >>> active_links((3, 4))
    array([ 1,  2,  5,  6, 11, 12, 13])

    If all the perimeter nodes `CLOSED_BOUNDARY` nodes, the only active link is between the two core nodes.

    >>> node_status = np.ones(3 * 4) * CLOSED_BOUNDARY
    >>> node_status[5:7] = CORE_NODE
    >>> active_links((3, 4), node_status_array=node_status)
    array([12])

    You can also provide a list of all the *from_nodes* and *to_nodes* for the grid. The following describes
    a grid with only a single link (between nodes 5 and 6).

    >>> active_links((3, 4), link_nodes=(np.array([5]), np.array([6])))
    array([0])
    """
    if node_status_array is None:
        node_status_array = node_status(shape)

    if link_nodes is None:
        (link_from_node, link_to_node) = node_index_at_link_ends(shape)
    else:
        (link_from_node, link_to_node) = link_nodes

    from_node_status = node_status_array[link_from_node]
    to_node_status = node_status_array[link_to_node]

    active_links = (((from_node_status == CORE_NODE) & ~
                     (to_node_status == CLOSED_BOUNDARY)) |
                    ((to_node_status == CORE_NODE) & ~
                     (from_node_status == CLOSED_BOUNDARY)))

    (active_links, ) = np.where(active_links)

    return active_links


def active_face_index(shape):
    return np.arange(active_face_count(shape))


def inlinks(shape, tonodes=None, return_count=True):
    links = np.vstack((south_links(shape), west_links(shape)))
    links.shape = (2, node_count(shape))
    return links


def outlinks(shape, tonodes=None, return_count=True):
    links = np.vstack((north_links(shape), east_links(shape)))
    links.shape = (2, node_count(shape))
    return links


def active_inlinks(shape, node_status=None, return_count=True):
    links = np.vstack((active_south_links(shape, node_status),
                       active_west_links(shape, node_status)))
    links.shape = (2, node_count(shape))
    return links


def active_outlinks(shape, node_status=None, return_count=True):
    links = np.vstack((active_north_links(shape, node_status),
                       active_east_links(shape, node_status)))
    links.shape = (2, node_count(shape))
    return links


def vertical_link_ids(shape):
    link_ids = np.arange(0, vertical_link_count(shape), dtype=np.int)
    link_ids.shape = (shape[0] - 1, shape[1])
    return link_ids


def horizontal_link_ids(shape):
    link_id_offset = vertical_link_count(shape)
    link_ids = np.arange(link_id_offset,
                         link_id_offset + horizontal_link_count(shape),
                         dtype=np.int)
    link_ids.shape = (shape[0], shape[1] - 1)
    return link_ids


def vertical_active_link_count(shape, node_status=None):
    if node_status is not None:
        is_inactive = (node_status == 0)
        is_inactive.shape = shape

        inactive_outlinks = is_inactive[:-1, 1:-1]
        inactive_inlinks = is_inactive[1:, 1:-1]
        inactive_links = inactive_outlinks | inactive_inlinks

        return (shape[0] - 1) * (shape[1] - 2) - np.sum(inactive_links.flat)
    else:
        return (shape[0] - 1) * (shape[1] - 2)


def horizontal_active_link_count(shape, node_status=None):
    if node_status is not None:
        is_inactive = (node_status == 0)
        is_inactive.shape = shape

        inactive_outlinks = is_inactive[1:-1, :-1]
        inactive_inlinks = is_inactive[1:-1, 1:]
        inactive_links = inactive_outlinks | inactive_inlinks

        return (shape[0] - 2) * (shape[1] - 1) - np.sum(inactive_links.flat)
    else:
        return (shape[0] - 2) * (shape[1] - 1)


def vertical_inactive_link_mask(shape, node_status):
    is_inactive = (node_status == 0)
    is_inactive.shape = shape

    inactive_outlinks = is_inactive[:-1, 1:-1]
    inactive_inlinks = is_inactive[1:, 1:-1]
    return inactive_outlinks | inactive_inlinks


def horizontal_inactive_link_mask(shape, node_status):
    is_inactive = (node_status == 0)
    is_inactive.shape = shape

    inactive_outlinks = is_inactive[1:-1, :-1]
    inactive_inlinks = is_inactive[1:-1, 1:]
    return inactive_outlinks | inactive_inlinks


#def vertical_active_link_ids(shape):
#    link_ids = np.arange(vertical_active_link_count(shape), dtype=np.int)
#    link_ids.shape = (shape[0] - 1, shape[1] - 2)
#    return link_ids


def vertical_active_link_ids(shape, node_status=None):
    if node_status is None:
        link_ids = np.arange(vertical_active_link_count(shape), dtype=np.int)
        #link_ids.shape = (shape[0] - 1, shape[1] - 2)
        #return link_ids
    else:
        inactive_links = vertical_inactive_link_mask(shape, node_status)
        inactive_links.shape = (inactive_links.size, )
        active_link_count = inactive_links.size - np.sum(inactive_links)

        link_ids = np.empty(inactive_links.size)
        link_ids[inactive_links] = - 1
        link_ids[~ inactive_links] = np.arange(active_link_count, dtype=int)

    link_ids.shape = (shape[0] - 1, shape[1] - 2)
    return link_ids


def horizontal_active_link_ids(shape, node_status=None):
    if node_status is None:
        link_id_offset = vertical_active_link_count(shape)
        link_ids = np.arange(
            link_id_offset,
            link_id_offset + horizontal_active_link_count(shape),
            dtype=np.int)
    else:
        link_id_offset = vertical_active_link_count(shape,
                                                    node_status=node_status)
        inactive_links = horizontal_inactive_link_mask(shape, node_status)
        inactive_links.shape = (inactive_links.size, )
        active_link_count = inactive_links.size - np.sum(inactive_links)

        link_ids = np.empty(inactive_links.size)
        link_ids[inactive_links] = - 1
        link_ids[~ inactive_links] = np.arange(
            link_id_offset, link_id_offset + active_link_count,
            dtype=np.int)
    link_ids.shape = (shape[0] - 2, shape[1] - 1)
    return link_ids


def west_links(shape):
    """
    >>> from landlab.utils.structured_grid import west_links
    >>> west_links((3, 4))
    array([[-1,  8,  9, 10],
           [-1, 11, 12, 13],
           [-1, 14, 15, 16]])
    """
    link_ids = horizontal_link_ids(shape)
    link_ids.shape = (shape[0], shape[1] - 1)
    return np.hstack((- np.ones((shape[0], 1), dtype=np.int), link_ids))


def north_links(shape):
    """
    >>> from landlab.utils.structured_grid import north_links
    >>> north_links((3, 4))
    array([[ 0,  1,  2,  3],
           [ 4,  5,  6,  7],
           [-1, -1, -1, -1]])
    """
    link_ids = vertical_link_ids(shape)
    link_ids.shape = (shape[0] - 1, shape[1])
    return np.vstack((link_ids, - np.ones((1, shape[1]), dtype=np.int)))


def south_links(shape):
    """
    >>> from landlab.utils.structured_grid import south_links
    >>> south_links((3, 4))
    array([[-1, -1, -1, -1],
           [ 0,  1,  2,  3],
           [ 4,  5,  6,  7]])
    """
    link_ids = vertical_link_ids(shape)
    link_ids.shape = (shape[0] - 1, shape[1])
    return np.vstack((- np.ones((1, shape[1]), dtype=np.int), link_ids))

def east_links(shape):
    """
    >>> from landlab.utils.structured_grid import east_links
    >>> east_links((3, 4))
    array([[ 8,  9, 10, -1],
           [11, 12, 13, -1],
           [14, 15, 16, -1]])
    """
    link_ids = horizontal_link_ids(shape)
    link_ids.shape = (shape[0], shape[1] - 1)
    return np.hstack((link_ids, - np.ones((shape[0], 1), dtype=int)))


def active_north_links(shape, node_status=None):
    """
    >>> from landlab.utils.structured_grid import active_north_links
    >>> active_north_links((3, 4))
    array([[-1,  0,  1, -1],
           [-1,  2,  3, -1],
           [-1, -1, -1, -1]])
    """
    active_north_links = np.empty(shape, dtype=int)
    try:
        links = vertical_active_link_ids(shape, node_status=node_status)
    except ValueError:
        pass
    links.shape = (shape[0] - 1, shape[1] - 2)
    active_north_links[:-1, 1:-1] = links
    active_north_links[:, (0, -1)] = -1
    active_north_links[-1, :] = -1

    return active_north_links


def active_south_links(shape, node_status=None):
    """
    >>> from landlab.utils.structured_grid import active_south_links
    >>> active_south_links((3, 4))
    array([[-1, -1, -1, -1],
           [-1,  0,  1, -1],
           [-1,  2,  3, -1]])
    """
    active_south_links = np.empty(shape, dtype=int)
    links = vertical_active_link_ids(shape, node_status=node_status)
    links.shape = (shape[0] - 1, shape[1] - 2)
    active_south_links[1:, 1:-1] = links
    active_south_links[:, (0, -1)] = -1
    active_south_links[0, :] = -1

    return active_south_links


def active_west_links(shape, node_status=None):
    """
    >>> from landlab.utils.structured_grid import active_west_links
    >>> active_west_links((3, 4))
    array([[-1, -1, -1, -1],
           [-1,  4,  5,  6],
           [-1, -1, -1, -1]])
    """
    active_west_links = np.empty(shape, dtype=int)
    try:
        active_west_links[1:-1, 1:] = horizontal_active_link_ids(
            shape, node_status=node_status)
    except ValueError:
        pass
    active_west_links[(0, -1), :] = -1
    active_west_links[:, 0] = -1

    return active_west_links


def active_east_links(shape, node_status=None):
    """
    >>> from landlab.utils.structured_grid import active_east_links
    >>> active_east_links((3, 4))
    array([[-1, -1, -1, -1],
           [ 4,  5,  6, -1],
           [-1, -1, -1, -1]])
    """
    active_east_links = np.empty(shape, dtype=int)
    active_east_links.fill(-999)
    try:
        active_east_links[1:-1, :-1] = horizontal_active_link_ids(
            shape, node_status=node_status)
    except ValueError:
        pass
    active_east_links[(0, -1), :] = -1
    active_east_links[:, -1] = -1

    return active_east_links


def outlink_count_per_node(shape):
    link_count = np.empty(shape, dtype=np.int)
    link_count[:-1, :-1] = 2
    link_count[-1, :-1] = 1
    link_count[:-1, -1] = 1
    link_count[-1, -1] = 0
    return np.ravel(link_count)


def inlink_count_per_node(shape):
    link_count = np.empty(shape, dtype=np.int)
    link_count[1:, 1:] = 2
    link_count[0, 1:] = 1
    link_count[1:, 0] = 1
    link_count[0, 0] = 0
    return np.ravel(link_count)


def active_outlink_count_per_node(shape):
    link_count = np.empty(shape, dtype=np.int)
    link_count[1:-1, 1:-1] = 2
    link_count[0, :] = 1
    link_count[-1, :] = 0
    link_count[:, 0] = 1
    link_count[:, -1] = 0

    link_count[0, 0] = 0
    link_count[-1, 0] = 0

    return np.ravel(link_count)


def active_inlink_count_per_node(shape):
    link_count = np.empty(shape, dtype=np.int)
    link_count[1:-1, 1:-1] = 2
    link_count[0, :] = 0
    link_count[-1, :] = 1
    link_count[:, 0] = 0
    link_count[:, -1] = 1

    link_count[0, -1] = 0
    link_count[-1, -1] = 0

    return np.ravel(link_count)


def setup_outlink_matrix(shape, return_count=True):
    links = outlinks(shape)
    if return_count:
        return (links, outlink_count_per_node(shape))
    else:
        return links


def setup_inlink_matrix(shape, return_count=True):
    links = inlinks(shape)
    if return_count:
        return (links, inlink_count_per_node(shape))
    else:
        return links


def setup_active_outlink_matrix(shape, node_status=None, return_count=True):
    links = active_outlinks(shape, node_status=node_status)
    if return_count:
        return links, active_outlink_count_per_node(shape)
    else:
        return links


def setup_active_inlink_matrix(shape, node_status=None, return_count=True):
    """Active links entering nodes.

    Return the IDs of the active links that enter each node of a grid. The shape of the returned array is
    (2, *N*) where *N* is the number of nodes in the grid. The first row contains the link ID entering the
    node from the bottom, and the second row the link entering the node from the left.

    Use the *return_count* keyword to, in addition to the link IDs, return the number of active links attached
    to each grid node.

    Use the *node_status_array* keyword to specify the status for each of the grid's nodes. If not given, each
    of the perimeter nodes is assumed to be `FIXED_VALUE_BOUNDARY`.

    Parameters
    ----------
    shape : tuple
        Shape of the structured grid
    node_status : array_like, optional
        Status of each node in the grid.
    return_count : boolean, optional
        If `True`, also return an array of active link counts per node.

    Returns
    -------
    links : (2, N) ndarray
        Active link IDs for each node.
    count : ndarray
        Number of active links per node.

    Examples
    --------
    Get the active link IDs for a grid of 3 nodes by 4 nodes. The first row list links entering nodes from the
    bottom, and the second links entering from the left.

    >>> from landlab.utils.structured_grid import setup_active_inlink_matrix
    >>> setup_active_inlink_matrix((3, 4), return_count=False)
    array([[-1, -1, -1, -1, -1,  0,  1, -1, -1,  2,  3, -1],
           [-1, -1, -1, -1, -1,  4,  5,  6, -1, -1, -1, -1]])
    >>> _, count = setup_active_inlink_matrix((3, 4))
    >>> count
    array([0, 0, 0, 0, 0, 2, 2, 1, 0, 1, 1, 0])
    """
    links = active_inlinks(shape, node_status=node_status)
    if return_count:
        return links, active_inlink_count_per_node(shape)
    else:
        return links


def node_index_with_halo(shape, halo_indices=BAD_INDEX_VALUE):
    """
    >>> from landlab.utils.structured_grid import node_index_with_halo
    >>> node_index_with_halo((2, 3), halo_indices=-1)
    array([[-1, -1, -1, -1, -1],
           [-1,  0,  1,  2, -1],
           [-1,  3,  4,  5, -1],
           [-1, -1, -1, -1, -1]])
    """
    shape_with_halo = np.array(shape) + 2

    ids = np.empty(shape_with_halo, dtype=np.int)

    (interiors, boundaries) = (interior_nodes(shape_with_halo),
                               boundary_nodes(shape_with_halo))

    ids.flat[interiors] = xrange(interior_node_count(shape_with_halo))
    ids.flat[boundaries] = halo_indices

    return ids


def cell_index_with_halo(shape, halo_indices=BAD_INDEX_VALUE,
                        inactive_indices=None):
    """
    >>> from landlab.utils.structured_grid import cell_index_with_halo
    >>> cell_index_with_halo((2, 3), halo_indices=-1)
    array([[-1, -1, -1, -1, -1],
           [-1,  0,  1,  2, -1],
           [-1,  3,  4,  5, -1],
           [-1, -1, -1, -1, -1]])

    >>> cell_index_with_halo((2, 3), halo_indices=-1, inactive_indices=-1)
    array([[-1, -1, -1, -1, -1],
           [-1, -1, -1, -1, -1],
           [-1, -1, -1, -1, -1],
           [-1, -1, -1, -1, -1]])
    """
    ids = node_index_with_halo(shape, halo_indices=halo_indices)
    if inactive_indices is not None:
        ids[:, (1, -2)] = inactive_indices
        ids[(1, -2), :] = inactive_indices

    return ids



def _neighbor_node_ids(ids_with_halo):
    shape = (ids_with_halo.shape[0] - 2, ids_with_halo.shape[1] - 2)
    kwds = {
        'strides': ids_with_halo.strides,
        'buffer': ids_with_halo,
        'dtype': ids_with_halo.dtype,
    }

    kwds['offset'] = ids_with_halo.itemsize * (ids_with_halo.shape[1])
    west_ids = np.ndarray(shape, **kwds)

    kwds['offset'] = ids_with_halo.itemsize * (ids_with_halo.shape[1] + 2)
    east_ids = np.ndarray(shape, **kwds)

    kwds['offset'] = ids_with_halo.itemsize
    south_ids = np.ndarray(shape, **kwds)

    kwds['offset'] = ids_with_halo.itemsize * (ids_with_halo.shape[1] * 2 + 1)
    north_ids = np.ndarray(shape, **kwds)

    return np.vstack(
        (east_ids.flat, north_ids.flat, west_ids.flat, south_ids.flat))


def _centered_node_ids(ids_with_halo):
    shape = (ids_with_halo.shape[0] - 2, ids_with_halo.shape[1] - 2)
    kwds = {'strides': ids_with_halo.strides,
            'buffer': ids_with_halo,
            'dtype': ids_with_halo.dtype, }

    kwds['offset'] = ids_with_halo.itemsize * (ids_with_halo.shape[1] + 1)
    return np.ndarray(shape, **kwds)


def neighbor_node_ids(shape, inactive=BAD_INDEX_VALUE):
    return linked_neighbor_node_ids(shape, [], inactive=inactive)


def linked_neighbor_node_ids(shape, closed_boundary_nodes,
                             open_boundary_nodes=None,
                             inactive=BAD_INDEX_VALUE):
    if open_boundary_nodes is None:
        open_boundary_nodes = []

    ids_with_halo = node_index_with_halo(shape, halo_indices=inactive)

    # Everything that touches a closed boundary is inactive
    if len(closed_boundary_nodes) > 0:
        ids = _centered_node_ids(ids_with_halo)
        ids.flat[closed_boundary_nodes] = inactive

    neighbors = _neighbor_node_ids(ids_with_halo)

    # Everything that a closed boundary touches is inactive
    if len(closed_boundary_nodes) > 0:
        neighbors[:, closed_boundary_nodes] = inactive

    if len(open_boundary_nodes) > 0:
        _set_open_boundary_neighbors(neighbors, open_boundary_nodes, inactive)

    return neighbors


def _set_open_boundary_neighbors(neighbors, open_boundary_nodes, value):
    open_boundary_neighbors = neighbors[:, open_boundary_nodes]
    is_open_boundary_neighbor = _find_open_boundary_neighbors(
        neighbors, open_boundary_nodes)
    nodes = np.choose(is_open_boundary_neighbor, (open_boundary_neighbors,
                                                  value))
    neighbors[:, open_boundary_nodes] = nodes


def _find_open_boundary_neighbors(neighbors, open_boundary_nodes):
    open_boundary_neighbors = neighbors[:, open_boundary_nodes]
    is_open_boundary_neighbor = np.in1d(open_boundary_neighbors,
                                        open_boundary_nodes)
    is_open_boundary_neighbor.shape = (neighbors.shape[0],
                                       len(open_boundary_nodes))
    return is_open_boundary_neighbor


def neighbor_node_array(shape, **kwds):
    """
    >>> from landlab.utils.structured_grid import neighbor_node_array
    >>> neighbors = neighbor_node_array((2, 3), inactive=-1)
    >>> neighbors.T
    array([[ 1,  3, -1, -1],
           [ 2,  4,  0, -1],
           [-1,  5,  1, -1],
           [ 4, -1, -1,  0],
           [ 5, -1,  3,  1],
           [-1, -1,  4,  2]])
    """
    closed_boundary_nodes = kwds.pop('closed_boundary_nodes', [])
    open_boundary_nodes = kwds.get('open_boundary_nodes', [])

    if len(closed_boundary_nodes) > 0 or len(open_boundary_nodes):
        neighbors = linked_neighbor_node_ids(shape, closed_boundary_nodes,
                                             **kwds)
    else:
        neighbors = neighbor_node_ids(shape, **kwds)

    return neighbors


def neighbor_cell_array(shape, out_of_bounds=BAD_INDEX_VALUE, contiguous=True):
    """
    >>> from landlab.utils.structured_grid import neighbor_cell_array
    >>> neighbors = neighbor_cell_array((2, 3), out_of_bounds=-1)
    >>> neighbors
    array([], dtype=int64)

    >>> neighbors = neighbor_cell_array((3, 3), out_of_bounds=-1)
    >>> neighbors
    array([[-1, -1, -1, -1]])

    >>> neighbors = neighbor_cell_array((5, 4), out_of_bounds=-1)
    >>> neighbors # doctest: +NORMALIZE_WHITESPACE
    array([[ 1,  2, -1, -1], [-1,  3,  0, -1],
           [ 3,  4, -1,  0], [-1,  5,  2,  1],
           [ 5, -1, -1,  2], [-1, -1,  4,  3]])
    """
    if cell_count(shape) > 0:
        shape = np.array(shape) - 2
        ids = node_index_with_halo(shape, halo_indices=out_of_bounds)

        neighbors = np.vstack((
            ids[1:shape[0] + 1, 2:].flat,
            ids[2:, 1:shape[1] + 1].flat,
            ids[1:shape[0] + 1, :shape[1]].flat,
            ids[:shape[0], 1:shape[1] + 1].flat,)).T
        if contiguous:
            return neighbors.copy()
        else:
            return neighbors
    else:
        return np.array([], dtype=np.int)


def diagonal_node_array(shape, out_of_bounds=BAD_INDEX_VALUE, contiguous=True,
                        boundary_node_mask=None):
    """
    Creates a list of IDs of the diagonal cells to each cell, as a 2D array.
    Only interior cells are assigned neighbors; boundary cells get -1 for
    each neighbor.  The order of the diagonal cells is [topright, topleft,
    bottomleft, bottomright].

    NG didn't touch this, but she thinks this should be nodes, not cells.

    >>> from landlab.utils.structured_grid import diagonal_node_array
    >>> diags = diagonal_node_array((2, 3), out_of_bounds=-1)
    >>> diags
    array([[ 4, -1, -1, -1],
           [ 5,  3, -1, -1],
           [-1,  4, -1, -1],
           [-1, -1, -1,  1],
           [-1, -1,  0,  2],
           [-1, -1,  1, -1]])
    >>> diags.flags['C_CONTIGUOUS']
    True
    >>> diags = diagonal_node_array((2, 3), out_of_bounds=-1, contiguous=False)
    >>> diags.flags['C_CONTIGUOUS']
    False
    """
    ids = node_index_with_halo(shape, halo_indices=out_of_bounds)

    diags = np.vstack((
        ids[2:, 2:].flat,
        ids[2:, :shape[1]].flat,
        ids[:shape[0], :shape[1]].flat,
        ids[:shape[0], 2:].flat,)).T

    if boundary_node_mask is not None:
        boundaries = np.empty(4, dtype=np.int)
        boundaries.fill(boundary_node_mask)
        diags[boundary_nodes(shape)] = boundaries

    if contiguous:
        return diags.copy()
    else:
        return diags


def diagonal_cell_array(shape, out_of_bounds=BAD_INDEX_VALUE, contiguous=True):
    """
    Construct a matrix of cell indices to each diagonally adjacent cell of a
    structured grid. If a cell does not have a diagonal neighbor, set the
    index for that neighbor to *out_of_bounds*.

    An grid without any cells returns an empty array.
    >>> from landlab.utils.structured_grid import diagonal_cell_array
    >>> diags = diagonal_cell_array((2, 3), out_of_bounds=-1)
    >>> diags
    array([], dtype=int64)

    A grid that has only one cell does not have any neighbors so all of its
    diagonals are set to *out_of_bounds*.
    >>> diags = diagonal_cell_array((3, 3), out_of_bounds=-1)
    >>> diags
    array([[-1, -1, -1, -1]])

    >>> diags = diagonal_cell_array((4, 4), out_of_bounds=-1)
    >>> diags # doctest: +NORMALIZE_WHITESPACE
    array([[ 3, -1, -1, -1], [-1,  2, -1, -1],
           [-1, -1, -1,  1], [-1, -1,  0, -1]])

    >>> diags = diagonal_cell_array((4, 5), out_of_bounds=-1)
    >>> diags # doctest: +NORMALIZE_WHITESPACE
    array([[ 4, -1, -1, -1], [ 5,  3, -1, -1], [-1,  4, -1, -1],
           [-1, -1, -1,  1], [-1, -1,  0,  2], [-1, -1,  1, -1]])
    """
    if cell_count(shape) > 0:
        shape = np.array(shape) - 2
        ids = node_index_with_halo(shape, halo_indices=out_of_bounds)

        diags = np.vstack((
            ids[2:, 2:].flat,
            ids[2:, :shape[1]].flat,
            ids[:shape[0], :shape[1]].flat,
            ids[:shape[0], 2:].flat,)).T
        if contiguous:
            return diags.copy()
        else:
            return diags
    else:
        return np.array([], dtype=np.int)


def diagonal_array_slow(shape, out_of_bounds=BAD_INDEX_VALUE):
    (nrows, ncols) = shape
    ncells = shape[0] * shape[1]
    diagonal_cells = - np.ones([ncells, 4], dtype=np.int)
    for r in xrange( 1, nrows-1 ):
        for c in xrange( 1, ncols-1 ):
            cell_id = r * ncols + c
            diagonal_cells[cell_id,2] = cell_id - ncols - 1 # bottom left
            diagonal_cells[cell_id,0] = cell_id + ncols + 1 # top right
            diagonal_cells[cell_id,3] = cell_id - ncols + 1 # bottom right
            diagonal_cells[cell_id,1] = cell_id + ncols - 1 # top left
    return diagonal_cells


def has_boundary_neighbor(neighbors, diagonals,
                          out_of_bounds=BAD_INDEX_VALUE):
    """
    DEJH thinks this method is broken since terminology update: it returns closed
    neihbors, not boundary neighbors.
    """
    return (out_of_bounds in neighbors |
            out_of_bounds in diagonals)


def has_boundary_neighbor_slow(neighbors, diagonals, out_of_bounds=BAD_INDEX_VALUE):
    # nbr_nodes=self.get_neighbor_list(id)
    # diag_nbrs=self.get_diagonal_list(id)

    i=0
    while i < 4 and neighbors[i] != out_of_bounds:
        i += 1

    if i < 4:
        return True
    else:
        r = 0
        while r < 4 and diagonals[r] != out_of_bounds:
            r += 1

    if r < 4 :
        return True
    else:
        return False


def reshape_array(shape, u, flip_vertically=False, copy=False):
    """
    >>> from landlab.utils.structured_grid import reshape_array
    >>> x = np.arange(12.)
    >>> y = reshape_array((3, 4), x)
    >>> y.shape
    (3, 4)
    >>> y
    array([[  0.,   1.,   2.,   3.],
           [  4.,   5.,   6.,   7.],
           [  8.,   9.,  10.,  11.]])
    >>> y.flags['C_CONTIGUOUS']
    True
    >>> x[0] = -1
    >>> y[0, 0]
    -1.0

    >>> x = np.arange(12.)
    >>> y = reshape_array((3, 4), x, flip_vertically=True)
    >>> y
    array([[  8.,   9.,  10.,  11.],
           [  4.,   5.,   6.,   7.],
           [  0.,   1.,   2.,   3.]])
    >>> y.flags['C_CONTIGUOUS']
    False
    >>> x[0] = -1
    >>> y[-1, 0]
    -1.0
    """
    reshaped_u = u.view()

    try:
        reshaped_u.shape = shape
    except ValueError:
        raise

    if flip_vertically:
        flipped_u = reshaped_u[::-1, :]
        if copy:
            return flipped_u.copy()
        else:
            return flipped_u
    else:
        if copy:
            return reshaped_u.copy()
        else:
            return reshaped_u

def nodes_around_points_on_unit_grid(shape, coords, mode='raise'):
    """
    Returns the nodes around a point on a structured grid with unit spacing
    and zero origin.

    >>> from landlab.utils.structured_grid import nodes_around_points_on_unit_grid
    >>> nodes_around_points_on_unit_grid((3, 3), (.1, .1))
    array([0, 3, 4, 1])

    >>> nodes_around_points_on_unit_grid((3, 3), (1., 1.))
    array([4, 7, 8, 5])
    """
    if isinstance(coords[0], np.ndarray):
        (rows, cols) = (coords[0].astype(np.int), coords[1].astype(np.int))
    else:
        (rows, cols) = (int(coords[0]), int(coords[1]))

    try:
        return np.ravel_multi_index(((rows, rows + 1, rows + 1, rows),
                                     (cols, cols, cols + 1, cols + 1)),
                                    shape, mode=mode).T
    except ValueError:
        raise

def nodes_around_points(shape, coords, spacing=(1., 1.),
                        origin=(0., 0.)):
    """
    Returns the nodes around a point on a structured grid with row and column
    *spacing*, and *origin*.

    >>> from landlab.utils.structured_grid import nodes_around_points
    >>> x = np.array([.9, 1.])
    >>> y = np.array([.1, 1.])
    >>> nodes_around_points((3, 3), (y, x))
    array([[0, 3, 4, 1],
           [4, 7, 8, 5]])

    >>> nodes_around_points((3, 3), (2., 1.))
    Traceback (most recent call last):
        ...
    ValueError: invalid entry in coordinates array
    """
    try:
        return nodes_around_points_on_unit_grid(
            shape,
            ((coords[0] - origin[0]) / spacing[0],
             (coords[1] - origin[1]) / spacing[1]))
    except ValueError:
        raise

def nodes_around_point(shape, coords, spacing=(1., 1.)):
    node_id = int(coords[0] // spacing[0] * shape[1] + coords[1] // spacing[1])
    if node_id + shape[1] + 1 >= shape[0] * shape[1] or node_id < 0:
        raise ValueError('invalid entry in coordinates array')

    return np.array([node_id, node_id + shape[1], node_id + shape[1] + 1,
                     node_id + 1])

def interior_node_id_to_node_id(shape, core_node_ids):
    """
    Converts the id of an interior node ID (i.e., if just the interior nodes
    were numbered) to a node ID.
    """
    IGW = shape[1]-2
    real_ID = (core_node_ids//IGW + 1) * shape[1] + (core_node_ids%IGW) + 1
    assert np.all(real_ID < shape[0]*shape[1])
    return real_ID.astype(int)

def node_id_to_interior_node_id(shape, node_ids):
    """
    Converts a node ID to the id of an interior node ID (i.e., if just the
    interior nodes were numbered)
    """
    ncols = shape[1]
    interior_ID = (node_ids//ncols - 1)*(ncols-2) + (node_ids%ncols) - 1
    if np.any(interior_ID < 0) or np.any(interior_ID >= (shape[0]-2)*(shape[1]-2)):
        print "One of the supplied nodes was outside the interior grid!"
        raise NameError()
    else:
        return interior_ID.astype(int)



import itertools

import numpy as np

from ..base import CORE_NODE, FIXED_VALUE_BOUNDARY


def number_of_nodes(shape):
    """Number of nodes is a structured quad grid.

    Parameters
    ----------
    shape : tuple of int
        Shape of grid of nodes.

    Returns
    -------
    int :
        Number of nodes in the grid.

    Examples
    --------
    >>> from landlab.grid.structured_quad.nodes import number_of_nodes
    >>> number_of_nodes((3, 4))
    12
    """
    return np.prod(shape, dtype=np.int)


def number_of_core_nodes(shape):
    """Number of core nodes is a structured quad grid.

    Parameters
    ----------
    shape : tuple of int
        Shape of grid of nodes.

    Returns
    -------
    int :
        Number of core nodes in the grid.

    Examples
    --------
    >>> from landlab.grid.structured_quad.nodes import number_of_core_nodes
    >>> number_of_core_nodes((3, 4))
    2
    """
    return np.prod(np.array(shape) - 2, dtype=np.int)


def corners(shape):
    """IDs of corner nodes.

    Parameters
    ----------
    shape : tuple of int
        Shape of grid of nodes.

    Returns
    -------
    (4, ) ndarray :
        IDs of the corner nodes.

    Examples
    --------
    >>> from landlab.grid.structured_quad.nodes import corners
    >>> corners((3, 4))
    array([ 0,  3,  8, 11])
    """
    node_count = number_of_nodes(shape)
    return np.array([0, shape[1] - 1, node_count - shape[1], node_count - 1],
                    dtype=np.int)


def node_ids(shape):
    """IDs of nodes.

    Parameters
    ----------
    shape : tuple of int
        Shape of grid of nodes.

    Returns
    -------
    ndarray :
        IDs of the nodes.

    Examples
    --------
    >>> from landlab.grid.structured_quad.nodes import node_ids
    >>> node_ids((3, 4))
    array([[ 0,  1,  2,  3],
           [ 4,  5,  6,  7],
           [ 8,  9, 10, 11]])
    """
    return np.arange(number_of_nodes(shape), dtype=np.int).reshape(shape)


def interior_nodes(shape):
    """IDs of interior nodes.

    Parameters
    ----------
    shape : tuple of int
        Shape of grid of nodes.

    Returns
    -------
    (M, N) ndarray :
        IDs of the interior nodes.

    Examples
    --------
    >>> from landlab.grid.structured_quad.nodes import interior_nodes
    >>> interior_nodes((3, 4))
    array([[5, 6]])
    >>> interior_nodes((4, 5))
    array([[ 6,  7,  8],
           [11, 12, 13]])
    """
    node_ids = np.arange(number_of_nodes(shape), dtype=np.int).reshape(shape)
    return node_ids[1:-1, 1:-1]


def top_iter(shape):
    """Iterator for the top perimeter nodes.

    Parameters
    ----------
    shape : tuple of int
        Shape of grid of nodes.
    """
    return xrange(shape[1] * (shape[0] - 1), shape[0] * shape[1])


def bottom_iter(shape):
    """Iterator for the bottom perimeter nodes.

    Parameters
    ----------
    shape : tuple of int
        Shape of grid of nodes.
    """
    return xrange(0, shape[1])


def left_iter(shape):
    """Iterator for the left perimeter nodes.

    Parameters
    ----------
    shape : tuple of int
        Shape of grid of nodes.
    """
    return xrange(0, shape[0] * shape[1], shape[1])


def right_iter(shape):
    """Iterator for the right perimeter nodes.

    Parameters
    ----------
    shape : tuple of int
        Shape of grid of nodes.
    """
    return xrange(shape[1] - 1, shape[0] * shape[1], shape[1])


def left_right_iter(shape, *args):
    """left_right_iter(shape, stop)
    left_right_iter(shape, start, stop[, step])

    Iterator for left and right perimeter nodes.

    Parameters
    ----------
    shape : tuple of int
        Shape of grid of nodes.
    start : int, optional
        Start row.
    stop : int, optional
        Stop row.
    step : int, optional
        Interval between rows.

    Examples
    --------
    >>> import numpy as np
    >>> from landlab.grid.structured_quad.nodes import left_right_iter
    >>> np.fromiter(left_right_iter((4, 3)), dtype=np.int)
    array([ 0,  2,  3,  5,  6,  8,  9, 11])
    >>> np.fromiter(left_right_iter((4, 3), 2), dtype=np.int)
    array([0, 2, 3, 5])
    >>> np.fromiter(left_right_iter((4, 3), 2, 4), dtype=np.int)
    array([ 6,  8,  9, 11])
    >>> np.fromiter(left_right_iter((4, 3), 1, 4, 2), dtype=np.int)
    array([ 3,  5,  9, 11])
    """
    if len(args) == 0:
        iter_rows = xrange(0, shape[0], 1)
    elif len(args) == 1:
        iter_rows = xrange(0, args[0], 1)
    elif len(args) == 2:
        iter_rows = xrange(args[0], args[1], 1)
    elif len(args) == 3:
        iter_rows = xrange(args[0], args[1], args[2])

    for row in iter_rows:
        yield row * shape[1]
        yield row * shape[1] + shape[1] - 1


def bottom_top_iter(shape):
    """Iterator for the bottom and top perimeter nodes.

    Parameters
    ----------
    shape : tuple of int
        Shape of grid of nodes.

    Examples
    --------
    >>> import numpy as np
    >>> from landlab.grid.structured_quad.nodes import bottom_top_iter
    >>> np.fromiter(bottom_top_iter((4, 3)), dtype=np.int)
    array([ 0,  1,  2,  9, 10, 11])
    """
    return itertools.chain(bottom_iter(shape), top_iter(shape))


def perimeter_iter(shape):
    """Iterator for all perimeter nodes.

    Parameters
    ----------
    shape : tuple of int
        Shape of grid of nodes.

    Examples
    --------
    >>> import numpy as np
    >>> from landlab.grid.structured_quad.nodes import perimeter_iter
    >>> np.fromiter(perimeter_iter((4, 3)), dtype=np.int)
    array([ 0,  1,  2,  3,  5,  6,  8,  9, 10, 11])
    """
    return itertools.chain(bottom_iter(shape),
                           left_right_iter(shape, 1, shape[0] - 1),
                           top_iter(shape))


def perimeter(shape):
    """Perimeter nodes.

    Parameters
    ----------
    shape : tuple of int
        Shape of grid of nodes.

    Returns
    -------
    ndarray :
        IDs of the perimeter nodes.

    Examples
    --------
    >>> from landlab.grid.structured_quad.nodes import perimeter
    >>> perimeter((3, 4))
    array([ 0,  1,  2,  3,  4,  7,  8,  9, 10, 11])

    """
    return np.fromiter(perimeter_iter(shape), dtype=np.int)


def status_with_perimeter_as_boundary(shape,
                                      node_status=FIXED_VALUE_BOUNDARY):
    """Node status for a grid whose boundary is along its perimeter.

    Parameters
    ----------
    shape : tuple of int
        Shape of grid of nodes.
    node_status : int or array_lik of int, optional
        Status of nodes on grid perimeter.

    Returns
    -------
    ndarray :
        Node status for grid.

    Examples
    --------
    >>> from landlab.grid.structured_quad.nodes import status_with_perimeter_as_boundary
    >>> status_with_perimeter_as_boundary((3, 4))
    array([[1, 1, 1, 1],
           [1, 0, 0, 1],
           [1, 1, 1, 1]])
    >>> status_with_perimeter_as_boundary((3, 4), node_status=-1)
    array([[-1, -1, -1, -1],
           [-1,  0,  0, -1],
           [-1, -1, -1, -1]])
    """
    status = np.empty(shape, dtype=int)
    status.fill(CORE_NODE)
    status.flat[perimeter(shape)] = node_status

    return status



#! /bin/env python

from landlab.io.vtk.encoders import AsciiEncoder, RawEncoder, Base64Encoder


class VtkEndian(object):
    def __init__ (self, endian):
        self.name = endian

    def __str__ (self):
        return self.name

    def __repr__(self):
        return 'VtkEndian(%s)' % self.name


class VtkType(object):
    def __init__(self, name, size):
        self.bytes = size
        self.name = name

    def __str__(self):
        return self.name

    def __repr__(self):
        return 'VtkType(%s, %s)' % (self.name, self.bytes)


class VtkCellType(object):
    def __init__(self, name, id):
        self.id = id
        self.name = name

    def __str__(self):
        return self.name

    def __int__(self):
        return self.id

    def __repr__(self):
        return 'VtkCellType(%s, %s)' % (self.name, self.id)


class VtkGridType(object):
    def __init__(self, name):
        self.name = name

    def __str__(self):
        return self.name

    def __repr__(self):
        return 'VtkGridType(%s)' % self.name


VtkLittleEndian = VtkEndian('LittleEndian')
VtkBigEndian = VtkEndian('BigEndian')


SYS_TO_VTK_ENDIAN = {
    'little': VtkLittleEndian,
    'big': VtkBigEndian
}


VtkUInt8 = VtkType('UInt8', 1)
VtkInt32 = VtkType('Int32', 4)
VtkInt64 = VtkType('Int64', 8)
VtkFloat32 = VtkType('Float32', 4)
VtkFloat64 = VtkType('Float64', 8)


NUMPY_TO_VTK_TYPE = {
    'uint8': VtkUInt8,
    'int32': VtkInt32,
    'int64': VtkInt64,
    'float32': VtkFloat32,
    'float64': VtkFloat64,
}

VTK_TO_NUMPY_TYPE = {
    'UInt8': 'uint8',
    'Int32': 'int32',
    'Int64': 'int64',
    'Float32': 'float32',
    'Float64': 'float64',
}


VtkVertex = VtkCellType('Vertex', 1)
VtkLine = VtkCellType('Line', 3)
VtkTriangle = VtkCellType('Triangle', 5)
VtkQuad = VtkCellType('Quad', 9)
VtkPolygon = VtkCellType('Polygon', 7)


EDGE_COUNT_TO_TYPE = {
    1: VtkVertex,
    2: VtkLine,
    3: VtkTriangle,
    4: VtkQuad,
}


VtkUniformRectilinear = VtkGridType('ImageData')
VtkRectilinear = VtkGridType('RectilinearGrid')
VtkStructured = VtkGridType('StructuredGrid')
VtkUnstructured = VtkGridType('UnstructuredGrid')


ENCODERS = {
    'ascii': AsciiEncoder(),
    'raw': RawEncoder(),
    'base64': Base64Encoder(),
}



import numpy
import scipy.sparse as sparse
import scipy.sparse.linalg as linalg
from landlab.grid.base import BAD_INDEX_VALUE

#these ones only so we can run this module ad-hoc:
#import pylab
from landlab import ModelParameterDictionary
#from copy import copy

#Things to add: 1. Explicit stability check. 2. Implicit handling of scenarios where kappa*dt exceeds critical step - subdivide dt automatically.

class PerronNLDiffuse(object):
    '''
    This module uses Taylor Perron's implicit (2011) method to solve the 
    nonlinear hillslope diffusion equation across a rectangular, regular grid
    for a single timestep. Note it works with the mass flux implicitly, and 
    thus does not actually calculate it. Grid must be at least 5x5.
    Built DEJH early June 2013.
    
    Boundary condition handling assumes each edge uses the same BC for each of 
    its nodes.
    This component cannot yet handle looped boundary conditions, but all others
    should be fine.
    
    This components requires the following parameters be set in the input file,
    *input_stream*, set in the component initialization:
        
        'uplift' or 'uplift_rate', both equivalent to the uplift rate
        'rock_density'
        'sed_density'
        'kappa', the diffusivity to use
        'S_crit', the maximum possible surface slope (radians)
    
    Optional inputs are:
        
        'dt', the model timestep (assumed constant)
        'values_to_diffuse', a string giving the name of the grid field
        containing the data to diffuse.
    
    If 'dt' is not supplied, you must call the method :func:`set_timestep` as
    part of your run loop. This allows you to set a dynamic timestep for this
    class.
    If 'values_to_diffuse' is not provided, defaults to 
    'topographic_elevation'.
    
    No particular units are necessary where they are not specified, as long as
    all units are internally consistent.
    
    The component takes *grid*, the RasterModelGrid object, and *input_stream*,
    the filename of (& optionally, path to) the parameter file, in its
    initialization.
    
    The primary method of this class is :func:`diffuse`.
    '''
    def __init__(self, grid, input_stream):
        inputs = ModelParameterDictionary(input_stream)
        self.inputs = inputs
        self.grid = grid
        
        self.internal_uplifts = True
                
        try:
            self._uplift = inputs.read_float('uplift')
        except:
            self._uplift = inputs.read_float('uplift_rate')
        self._rock_density = inputs.read_float('rock_density')
        self._sed_density = inputs.read_float('sed_density')
        self._kappa = inputs.read_float('kappa') # ==_a
        self._S_crit = inputs.read_float('S_crit')
        try:
            self.values_to_diffuse = inputs.read_str('values_to_diffuse')
        except:
            self.values_to_diffuse = 'topographic_elevation'
        try:
            self.timestep_in = inputs.read_float('dt')  
        except:
            print 'No fixed timestep supplied, it must be set dynamically somewhere else. Be sure to call input_timestep(timestep_in) as part of your run loop.'
        
        self._delta_x = grid.node_spacing
        self._delta_y = self._delta_x
        self._one_over_delta_x = 1./self._delta_x
        self._one_over_delta_y = 1./self._delta_y
        self._one_over_delta_x_sqd = self._one_over_delta_x**2.
        self._one_over_delta_y_sqd = self._one_over_delta_y**2.
        self._b = 1./self._S_crit**2.
        
        ncols = grid.number_of_node_columns
        self.ncols = ncols
        nrows = grid.number_of_node_rows
        self.nrows = nrows
        nnodes = grid.number_of_nodes
        self.nnodes = nnodes
        ninteriornodes = grid.number_of_interior_nodes
        ncorenodes = ninteriornodes-2*(ncols+nrows-6)
        self.ninteriornodes = ninteriornodes
        self.interior_grid_width = ncols-2
        self.core_cell_width = ncols-4
    
        self._interior_corners = numpy.array([ncols+1,2*ncols-2,nnodes-2*ncols+1,nnodes-ncols-2])
        _left_list = numpy.array(range(2*ncols+1,nnodes-2*ncols,ncols)) #these are still real IDs
        _right_list = numpy.array(range(3*ncols-2,nnodes-2*ncols,ncols))
        _bottom_list = numpy.array(range(ncols+2,2*ncols-2))
        _top_list = numpy.array(range(nnodes-2*ncols+2,nnodes-ncols-2))
        self._left_list = _left_list
        self._right_list = _right_list
        self._bottom_list = _bottom_list
        self._top_list = _top_list

        self._core_nodes = self.coreIDtoreal(numpy.arange(ncorenodes,dtype=int))
        self.corenodesbyintIDs = self.realIDtointerior(self._core_nodes)
        self.ncorenodes = len(self._core_nodes)
        
        self.corner_interior_IDs = self.realIDtointerior(self._interior_corners) #i.e., interior corners as interior IDs
        self.bottom_interior_IDs = self.realIDtointerior(numpy.array(_bottom_list))
        self.top_interior_IDs = self.realIDtointerior(numpy.array(_top_list))
        self.left_interior_IDs = self.realIDtointerior(numpy.array(_left_list))
        self.right_interior_IDs = self.realIDtointerior(numpy.array(_right_list))
        
        #build an ID map to let us easily map the variables of the core nodes onto the operating matrix:
        #This array is ninteriornodes long, but the IDs it contains are REAL IDs
        operating_matrix_ID_map = numpy.empty((ninteriornodes,9))
        self.interior_IDs_as_real = self.interiorIDtoreal(numpy.arange(ninteriornodes))
        for j in xrange(ninteriornodes):
            i = self.interior_IDs_as_real[j]
            #operating_matrix_ID_map[:,j] = numpy.array([(i-ncols+1),(i-ncols+2),(i-ncols+3),(i-1),i,(i+1),(i+ncols-3),(i+ncols-2),(i+ncols-1)])
            operating_matrix_ID_map[j,:] = numpy.array([(i-ncols-1),(i-ncols),(i-ncols+1),(i-1),i,(i+1),(i+ncols-1),(i+ncols),(i+ncols+1)])
        self.operating_matrix_ID_map = operating_matrix_ID_map
        self.operating_matrix_core_int_IDs = self.realIDtointerior(operating_matrix_ID_map[self.corenodesbyintIDs,:]) #shape(ncorenodes,9)
        #see below for corner and edge maps
        
        #Build masks for the edges and corners to be applied to the operating matrix map.
        #Antimasks are the boundary nodes, masks are "normal"
        topleft_mask = [1,2,4,5]
        topleft_antimask = [0,3,6,7,8]
        topright_mask = [0,1,3,4]
        topright_antimask = [2,5,6,7,8]
        bottomleft_mask = [4,5,7,8]
        bottomleft_antimask = [0,1,2,3,6]
        bottomright_mask = [3,4,6,7]
        bottomright_antimask = [0,1,2,5,8]
        self.corners_masks = (numpy.vstack((bottomleft_mask,bottomright_mask,topleft_mask,topright_mask))) #(each_corner,mask_for_each_corner)
        self.corners_antimasks = (numpy.vstack((bottomleft_antimask,bottomright_antimask,topleft_antimask,topright_antimask))) #so shape becomes (4,5)
        self.left_mask = [1,2,4,5,7,8]
        self.left_antimask = [0,3,6]
        self.top_mask = [0,1,2,3,4,5]
        self.top_antimask = [6,7,8]
        self.right_mask = [0,1,3,4,6,7]
        self.right_antimask = [2,5,8]
        self.bottom_mask = [3,4,5,6,7,8]
        self.bottom_antimask = [0,1,2]
        self.antimask_corner_position = [0,2,2,4] #this is the position w/i the corner antimasks that the true corner actually occupies
        
        self.modulator_mask = numpy.array([-ncols-1,-ncols,-ncols+1,-1,0,1,ncols-1,ncols,ncols+1])
        
        #Set up terms for BC handling (still feels very clumsy)
        bottom_nodes = grid.bottom_edge_node_ids()
        top_nodes = grid.top_edge_node_ids()
        left_nodes = grid.left_edge_node_ids()
        right_nodes = grid.right_edge_node_ids()
        self.bottom_flag = 1
        self.top_flag = 1
        self.left_flag = 1
        self.right_flag = 1
        #self.corner_flags = [1,1,1,1] #In ID order, so BL,BR,TL,TR
        if numpy.all(grid.node_status[bottom_nodes[1:-1]]==4): #This should be all of them, or none of them
            self.bottom_flag = 4
        elif numpy.all(grid.node_status[bottom_nodes[1:-1]]==3):
            self.bottom_flag = 3
        elif numpy.all(grid.node_status[bottom_nodes[1:-1]]==2):
            self.bottom_flag = 2
        elif numpy.all(grid.node_status[bottom_nodes[1:-1]]==1):
            pass
        else:
            raise NameError('Different cells on the same grid edge have different boundary statuses!!')
            #Note this could get fraught if we need to open a cell to let water flow out...
        if numpy.all(grid.node_status[top_nodes[1:-1]]==4):
            self.top_flag = 4
        elif numpy.all(grid.node_status[top_nodes[1:-1]]==3):
            self.top_flag = 3
        elif numpy.all(grid.node_status[top_nodes[1:-1]]==2):
            self.top_flag = 2
        elif numpy.all(grid.node_status[top_nodes[1:-1]]==1):
            pass
        else:
            raise NameError('Different cells on the same grid edge have different boundary statuses!!')
        if numpy.all(grid.node_status[left_nodes[1:-1]]==4):
            self.left_flag = 4
        elif numpy.all(grid.node_status[left_nodes[1:-1]]==3):
            self.left_flag = 3
        elif numpy.all(grid.node_status[left_nodes[1:-1]]==2):
            self.left_flag = 2
        elif numpy.all(grid.node_status[left_nodes[1:-1]]==1):
            pass
        else:
            raise NameError('Different cells on the same grid edge have different boundary statuses!!')
        if numpy.all(grid.node_status[right_nodes[1:-1]]==4):
            self.right_flag = 4	       
        elif numpy.all(grid.node_status[right_nodes[1:-1]]==3):
            self.right_flag = 3	       
        elif numpy.all(grid.node_status[right_nodes[1:-1]]==2):
            self.right_flag = 2
        elif numpy.all(grid.node_status[right_nodes[1:-1]]==1):
            pass
        else:
            raise NameError('Different cells on the same grid edge have different boundary statuses!!')
        
        self.fixed_grad_BCs_present = self.bottom_flag==2 or self.top_flag==2 or self.left_flag==2 or self.right_flag==2
        self.looped_BCs_present = self.bottom_flag==3 or self.top_flag==3 or self.left_flag==3 or self.right_flag==3
        if self.fixed_grad_BCs_present:
            if self.values_to_diffuse != grid.fixed_gradient_of:
                raise ValueError("Boundary conditions set in the grid don't apply to the data the diffuser is trying to work with!")
        
        if numpy.any(grid.node_status == 2):    
            self.fixed_grad_offset_map = numpy.empty(nrows*ncols, dtype = float)
            self.fixed_grad_anchor_map = numpy.empty_like(self.fixed_grad_offset_map)
            self.fixed_grad_offset_map[self.grid.fixed_gradient_node_properties['boundary_node_IDs']] = self.grid.fixed_gradient_node_properties['values_to_add']
            #self.fixed_grad_anchor_map[self.grid.fixed_gradient_node_properties['boundary_node_IDs']] = self.grid.fixed_gradient_node_properties['anchor_node_IDs'] #not actually necessary

        self.corner_flags = grid.node_status[[0,ncols-1,-ncols,-1]]
        
        op_mat_just_corners = operating_matrix_ID_map[self.corner_interior_IDs,:]
        op_mat_cnr0 = op_mat_just_corners[0,bottomleft_mask]
        op_mat_cnr1 = op_mat_just_corners[1,bottomright_mask]
        op_mat_cnr2 = op_mat_just_corners[2,topleft_mask]
        op_mat_cnr3 = op_mat_just_corners[3,topright_mask]
        op_mat_just_active_cnrs = numpy.vstack((op_mat_cnr0,op_mat_cnr1,op_mat_cnr2,op_mat_cnr3))
        #op_mat_anticnr0 = op_mat_just_corners[0,bottomleft_antimask]
        #op_mat_anticnr1 = op_mat_just_corners[1,bottomright_antimask]
        #op_mat_anticnr2 = op_mat_just_corners[2,topleft_antimask]
        #op_mat_anticnr3 = op_mat_just_corners[3,topright_antimask]
        #op_mat_just_active_anticnrs = numpy.vstack((op_mat_anticnr0,op_mat_anticnr1,op_mat_anticnr2,op_mat_anticnr3))
        self.operating_matrix_corner_int_IDs = self.realIDtointerior(op_mat_just_active_cnrs) #(4corners,4nodesactivepercorner)
        #self.operating_matrix_anticorner_int_IDs = self.realIDtointerior(op_mat_just_active_anticnrs) #(4corners,5nodesantipercorner)
        self.operating_matrix_bottom_int_IDs = self.realIDtointerior(operating_matrix_ID_map[self.bottom_interior_IDs,:][:,self.bottom_mask]) #(nbottomnodes,6activenodeseach)
        self.operating_matrix_top_int_IDs = self.realIDtointerior(operating_matrix_ID_map[self.top_interior_IDs,:][:,self.top_mask])
        self.operating_matrix_left_int_IDs = self.realIDtointerior(operating_matrix_ID_map[self.left_interior_IDs,:][:,self.left_mask])
        self.operating_matrix_right_int_IDs = self.realIDtointerior(operating_matrix_ID_map[self.right_interior_IDs,:][:,self.right_mask])
        print "setup complete"

    def input_timestep(self, timestep_in):
        """
        Allows the user to set a dynamic (evolving) timestep manually as part of
        a run loop.
        """
        self.timestep_in = timestep_in

    def gear_timestep(self, timestep_in, new_grid):
        """
        This method allows the gearing between the model run step and the 
        component (shorter) step.
        The method becomes unstable if S>Scrit, so we test to prevent this.
        We implicitly assume the initial condition does not contain 
        slopes > Scrit. If the method persistently explodes, this may be the 
        problem.
        """
        extended_elevs = numpy.empty(self.grid.number_of_nodes+1, dtype=float)
        extended_elevs[-1] = numpy.nan
        node_neighbors = self.grid.get_neighbor_list()
        extended_elevs[:-1] = new_grid['node'][self.values_to_diffuse]
        max_offset = numpy.nanmax(numpy.fabs(extended_elevs[:-1][node_neighbors] - extended_elevs[:-1].reshape((self.grid.number_of_nodes,1))))
        if max_offset > numpy.tan(self._S_crit) * self.grid.dx: #using S not tan(S) adds a buffer - but not appropriate
            self.internal_repeats = int(max_offset//(numpy.tan(self._S_crit)*self.grid.dx)) + 1
            #now we rig it so the actual timestep is an integer divisor of T_in:
            self._delta_t = timestep_in / self.internal_repeats
            self.uplift_per_step = (new_grid['node'][self.values_to_diffuse]-self.grid['node'][self.values_to_diffuse])/self.internal_repeats
            if self.internal_repeats > 10000:
                raise ValueError('Uplift rate is too high; solution is not stable!!')
            #print 'Reducing internal timestep..., internal loops=', self.internal_repeats
        else:
            self.internal_repeats = 1
            self._delta_t = timestep_in
            self.uplift_per_step = new_grid['node'][self.values_to_diffuse]-self.grid['node'][self.values_to_diffuse]
        return self._delta_t
        
    #@profile #for line_profiler
    def set_variables(self, grid):
        '''
        This function sets the variables needed for update().
        Now vectorized, shouold run faster.
        At the moment, this method can only handle fixed value BCs.
        '''
        n_interior_nodes = grid.number_of_interior_nodes
        ###_operating_matrix = sparse.lil_matrix((n_interior_nodes, n_interior_nodes), dtype=float)
        ####we're going to try this with COO, so initialize later

        #Initialize the local builder lists
        _mat_RHS = numpy.zeros(n_interior_nodes)
    
        try:
            elev = grid['node'][self.values_to_diffuse]
        except:
            print 'elevations not found in grid!'
        try:
            _delta_t = self._delta_t
        except:
            raise NameError('Timestep not set! Call gear_timestep(tstep) after initializing the component, but before running it.')
        _one_over_delta_x = self._one_over_delta_x
        _one_over_delta_x_sqd = self._one_over_delta_x_sqd
        _one_over_delta_y = self._one_over_delta_y
        _one_over_delta_y_sqd = self._one_over_delta_y_sqd
        _kappa = self._kappa
        _b = self._b
        _S_crit = self._S_crit
        _core_nodes = self._core_nodes
        corenodesbyintIDs = self.corenodesbyintIDs
        #interior_grid_width = self.interior_grid_width
        #core_cell_width = self.core_cell_width
        #operating_matrix_ID_map = self.operating_matrix_ID_map
        operating_matrix_core_int_IDs = self.operating_matrix_core_int_IDs
        operating_matrix_corner_int_IDs = self.operating_matrix_corner_int_IDs
        _interior_corners = self._interior_corners
        corners_antimasks = self.corners_antimasks
        corner_interior_IDs = self.corner_interior_IDs
        modulator_mask = self.modulator_mask
        corner_flags = self.corner_flags
        bottom_interior_IDs = self.bottom_interior_IDs
        top_interior_IDs = self.top_interior_IDs
        left_interior_IDs = self.left_interior_IDs
        right_interior_IDs = self.right_interior_IDs
        bottom_antimask = self.bottom_antimask
        _bottom_list = self._bottom_list
        top_antimask = self.top_antimask
        _top_list = self._top_list
        left_antimask = self.left_antimask
        _left_list = self._left_list
        right_antimask = self.right_antimask
        _right_list = self._right_list
        
        #Need to modify the "effective" values of the edge nodes if any of the edges are inactive:
        if self.bottom_flag == 4:
            bottom_nodes = grid.bottom_edge_node_ids()
            elev[bottom_nodes] = elev[bottom_nodes+grid.shape[1]]
            #corners are special cases, and assumed linked to the bottom and top edge BCs...
            elev[bottom_nodes[0]] = elev[bottom_nodes[0]+grid.shape[1]+1]
            elev[bottom_nodes[-1]] = elev[bottom_nodes[-1]+grid.shape[1]-1]
        if self.top_flag == 4:
            top_nodes = grid.top_edge_node_ids()
            elev[top_nodes] = elev[top_nodes-grid.shape[1]]
            #corners are special cases, and assumed linked to the bottom and top edge BCs...
            elev[top_nodes[0]] = elev[top_nodes[0]-grid.shape[1]+1]
            elev[top_nodes[-1]] = elev[top_nodes[-1]-grid.shape[1]-1]
        if self.left_flag ==4:
            left_nodes = grid.left_edge_node_ids()
            elev[left_nodes[1:-1]] = elev[left_nodes[1:-1]+1]
        if self.right_flag ==4:
            right_nodes = grid.right_edge_node_ids()
            elev[right_nodes[1:-1]] = elev[right_nodes[1:-1]-1]

        #replacing loop:
        cell_neighbors = grid.get_neighbor_list() #E,N,W,S
        cell_diagonals = grid.get_diagonal_list() #NE,NW,SW,SE
        cell_neighbors[cell_neighbors==BAD_INDEX_VALUE] = -1
        cell_diagonals[cell_diagonals==BAD_INDEX_VALUE] = -1
        
        _z_x = (elev[cell_neighbors[:,0]]-elev[cell_neighbors[:,2]])*0.5*_one_over_delta_x
        _z_y = (elev[cell_neighbors[:,1]]-elev[cell_neighbors[:,3]])*0.5*_one_over_delta_y
        _z_xx = (elev[cell_neighbors[:,0]]-2.*elev+elev[cell_neighbors[:,2]])*_one_over_delta_x_sqd
        _z_yy = (elev[cell_neighbors[:,1]]-2.*elev+elev[cell_neighbors[:,3]])*_one_over_delta_y_sqd
        _z_xy = (elev[cell_diagonals[:,0]] - elev[cell_diagonals[:,1]] - elev[cell_diagonals[:,3]] + elev[cell_diagonals[:,2]])*0.25*_one_over_delta_x*_one_over_delta_y
        _d = 1./(1.-_b*(_z_x*_z_x+_z_y*_z_y))
        
        _abd_sqd = _kappa*_b*_d*_d
        _F_ij = -2.*_kappa*_d*(_one_over_delta_x_sqd+_one_over_delta_y_sqd) - 4.*_abd_sqd*(_z_x*_z_x*_one_over_delta_x_sqd+_z_y*_z_y*_one_over_delta_y_sqd)
        _F_ijminus1 = _kappa*_d*_one_over_delta_x_sqd - _abd_sqd*_z_x*(_z_xx+_z_yy)*_one_over_delta_x - 4.*_abd_sqd*_b*_d*(_z_x*_z_x*_z_xx+_z_y*_z_y*_z_yy+2.*_z_x*_z_y*_z_xy)*_z_x*_one_over_delta_x - 2.*_abd_sqd*(_z_x*_z_xx*_one_over_delta_x-_z_x*_z_x*_one_over_delta_x_sqd+_z_y*_z_xy*_one_over_delta_x)
        _F_ijplus1 = _kappa*_d*_one_over_delta_x_sqd + _abd_sqd*_z_x*(_z_xx+_z_yy)*_one_over_delta_x + 4.*_abd_sqd*_b*_d*(_z_x*_z_x*_z_xx+_z_y*_z_y*_z_yy+2.*_z_x*_z_y*_z_xy)*_z_x*_one_over_delta_x + 2.*_abd_sqd*(_z_x*_z_xx*_one_over_delta_x+_z_x*_z_x*_one_over_delta_x_sqd+_z_y*_z_xy*_one_over_delta_x)
        _F_iminus1j = _kappa*_d*_one_over_delta_y_sqd - _abd_sqd*_z_y*(_z_xx+_z_yy)*_one_over_delta_y - 4.*_abd_sqd*_b*_d*(_z_x*_z_x*_z_xx+_z_y*_z_y*_z_yy+2.*_z_x*_z_y*_z_xy)*_z_y*_one_over_delta_y - 2.*_abd_sqd*(_z_y*_z_yy*_one_over_delta_y-_z_y*_z_y*_one_over_delta_y_sqd+_z_x*_z_xy*_one_over_delta_y)
        _F_iplus1j = _kappa*_d*_one_over_delta_y_sqd + _abd_sqd*_z_y*(_z_xx+_z_yy)*_one_over_delta_y + 4.*_abd_sqd*_b*_d*(_z_x*_z_x*_z_xx+_z_y*_z_y*_z_yy+2.*_z_x*_z_y*_z_xy)*_z_y*_one_over_delta_y + 2.*_abd_sqd*(_z_y*_z_yy*_one_over_delta_y+_z_y*_z_y*_one_over_delta_y_sqd+_z_x*_z_xy*_one_over_delta_y)
        _F_iplus1jplus1 = _abd_sqd*_z_x*_z_y*_one_over_delta_x*_one_over_delta_y
        _F_iminus1jminus1 = _F_iplus1jplus1
        _F_iplus1jminus1 = -_F_iplus1jplus1
        _F_iminus1jplus1 = _F_iplus1jminus1
        
        _equ_RHS_calc_frag = _F_ij*elev+_F_ijminus1*elev[cell_neighbors[:,2]]+_F_ijplus1*elev[cell_neighbors[:,0]]+_F_iminus1j*elev[cell_neighbors[:,3]]+_F_iplus1j*elev[cell_neighbors[:,1]]+_F_iminus1jminus1*elev[cell_diagonals[:,2]]+_F_iplus1jplus1*elev[cell_diagonals[:,0]]+_F_iplus1jminus1*elev[cell_diagonals[:,1]]+_F_iminus1jplus1*elev[cell_diagonals[:,3]]
        
        #NB- all _z_... and _F_... variables are nnodes long, and thus use real IDs (tho calcs will be flawed for Bnodes)
        
        #RHS of equ 6 (see para [20])
        _func_on_z = self._rock_density/self._sed_density*self._uplift + _kappa*((_z_xx+_z_yy)/(1.-(_z_x*_z_x+_z_y*_z_y)/_S_crit*_S_crit) + 2.*(_z_x*_z_x*_z_xx+_z_y*_z_y*_z_yy+2.*_z_x*_z_y*_z_xy)/(_S_crit*_S_crit*(1.-(_z_x*_z_x+_z_y*_z_y)/_S_crit*_S_crit)**2.))

        #Remember, the RHS is getting wiped each loop as part of self.set_variables()
        #_mat_RHS is ninteriornodes long, but were only working on a ncorenodes long subset here
        _mat_RHS[corenodesbyintIDs] += elev[_core_nodes] + _delta_t*(_func_on_z[_core_nodes] - _equ_RHS_calc_frag[_core_nodes])
        low_row = numpy.vstack((_F_iminus1jminus1, _F_iminus1j, _F_iminus1jplus1))*-_delta_t
        mid_row = numpy.vstack((-_delta_t*_F_ijminus1, 1.-_delta_t*_F_ij, -_delta_t*_F_ijplus1))
        top_row = numpy.vstack((_F_iplus1jminus1, _F_iplus1j, _F_iplus1jplus1))*-_delta_t
        nine_node_map = numpy.vstack((low_row,mid_row,top_row)).T #Note shape is (nnodes,9); it's realID indexed
        #_operating_matrix[(operating_matrix_core_int_IDs.astype(int),numpy.arange(9,dtype=int).reshape((1,9)))] += nine_node_map[_core_nodes,:] #is there something weird happening with the redimensionalizing here...?
        core_op_mat_row = numpy.repeat(corenodesbyintIDs, 9)
        core_op_mat_col = operating_matrix_core_int_IDs.astype(int).flatten()
        core_op_mat_data = nine_node_map[_core_nodes,:].flatten()
        ###_operating_matrix[(corenodesbyintIDs.reshape((self.ncorenodes,1)),operating_matrix_core_int_IDs.astype(int))] += nine_node_map[_core_nodes,:] #this should now be putting these values in the right cells...
        
        #Now the interior corners; BL,BR,TL,TR
        _mat_RHS[corner_interior_IDs] += elev[_interior_corners] + _delta_t*(_func_on_z[_interior_corners] - _equ_RHS_calc_frag[_interior_corners])
        corners_op_mat_row = numpy.repeat(self.corner_interior_IDs,4)
        corners_op_mat_col = operating_matrix_corner_int_IDs.astype(int).flatten()
        corners_op_mat_data = nine_node_map[_interior_corners,:][(numpy.arange(4).reshape((4,1)),self.corners_masks)].flatten() #1st index gives (4,9), 2nd reduces to (4,4), then flattened
        ###_operating_matrix[(self.corner_interior_IDs.reshape((4,1)),operating_matrix_corner_int_IDs.astype(int))] += nine_node_map[_interior_corners,:][(numpy.arange(4).reshape((4,1)),self.corners_masks)] #rhs 1st index gives (4,9), 2nd reduces to (4,4)
        for i in range(4): #loop over each corner, as so few
            #Note that this ONLY ADDS THE VALUES FOR THE TRUE GRID CORNERS. The sides get done in the edge tests, below.
            if corner_flags[i] == 1:
                true_corner = self.antimask_corner_position[i]
                _mat_RHS[corner_interior_IDs[i]] -= _delta_t*numpy.sum(nine_node_map[_interior_corners[i],:][corners_antimasks[i,true_corner]]*elev[_interior_corners[i]+modulator_mask[corners_antimasks[i,true_corner]]])
            elif corner_flags[i] == 4 or corner_flags[i] == 3: #inactive boundary cell
                #Actually the easiest case! Equivalent to fixed gradient, but the gradient is zero, so material only goes in the linked cell. And because it's a true corner, that linked cell doesn't appear in the interior matrix at all!
                pass
            elif corner_flags[i] == 2:
                true_corner = self.antimask_corner_position[i]
                _mat_RHS[corner_interior_IDs[i]] -= _delta_t*numpy.sum(nine_node_map[_interior_corners[i],:][corners_antimasks[i,true_corner]]*self.fixed_gradient_offset_map[_interior_corners[i]+modulator_mask[corners_antimasks[i,true_corner]]])
            else:
                raise NameError('Sorry! This module cannot yet handle fixed gradient or looped BCs...')
            #Todo: handle these BCs properly, once the grid itself works with them.
            #Can follow old routines; see self.set_bc_cell() commented out method below.
            
        #Now the edges
        _mat_RHS[bottom_interior_IDs] += elev[_bottom_list] + _delta_t*(_func_on_z[_bottom_list] - _equ_RHS_calc_frag[_bottom_list])
        _mat_RHS[top_interior_IDs] += elev[_top_list] + _delta_t*(_func_on_z[_top_list] - _equ_RHS_calc_frag[_top_list])
        _mat_RHS[left_interior_IDs] += elev[_left_list] + _delta_t*(_func_on_z[_left_list] - _equ_RHS_calc_frag[_left_list])        
        _mat_RHS[right_interior_IDs] += elev[_right_list] + _delta_t*(_func_on_z[_right_list] - _equ_RHS_calc_frag[_right_list])
        bottom_op_mat_row = numpy.repeat(bottom_interior_IDs,6)
        top_op_mat_row = numpy.repeat(top_interior_IDs,6)
        left_op_mat_row = numpy.repeat(left_interior_IDs,6)
        right_op_mat_row = numpy.repeat(right_interior_IDs,6)   
        bottom_op_mat_col = self.operating_matrix_bottom_int_IDs.astype(int).flatten()
        top_op_mat_col = self.operating_matrix_top_int_IDs.astype(int).flatten()
        left_op_mat_col = self.operating_matrix_left_int_IDs.astype(int).flatten()
        right_op_mat_col = self.operating_matrix_right_int_IDs.astype(int).flatten()
        bottom_op_mat_data = nine_node_map[_bottom_list,:][:,self.bottom_mask].flatten()
        top_op_mat_data = nine_node_map[_top_list,:][:,self.top_mask].flatten()
        left_op_mat_data = nine_node_map[_left_list,:][:,self.left_mask].flatten()
        right_op_mat_data = nine_node_map[_right_list,:][:,self.right_mask].flatten()
        ###_operating_matrix[bottom_interior_IDs.reshape(bottom_interior_IDs.size,1),self.operating_matrix_bottom_int_IDs.astype(int)] += nine_node_map[_bottom_list,:][:,self.bottom_mask]
        ###_operating_matrix[top_interior_IDs.reshape(top_interior_IDs.size,1),self.operating_matrix_top_int_IDs.astype(int)] += nine_node_map[_top_list,:][:,self.top_mask]
        ###_operating_matrix[left_interior_IDs.reshape(left_interior_IDs.size,1),self.operating_matrix_left_int_IDs.astype(int)] += nine_node_map[_left_list,:][:,self.left_mask]
        ###_operating_matrix[right_interior_IDs.reshape(right_interior_IDs.size,1),self.operating_matrix_right_int_IDs.astype(int)] += nine_node_map[_right_list,:][:,self.right_mask]
        
        if self.bottom_flag == 1:
            #goes to RHS only
            _mat_RHS[bottom_interior_IDs] -= _delta_t*numpy.sum(nine_node_map[_bottom_list,:][:,bottom_antimask]*elev[_bottom_list.reshape((len(_bottom_list),1))+(modulator_mask[bottom_antimask]).reshape(1,3)], axis=1) #note the broadcasting to (nedge,3) in final fancy index   
            #...& the corners
            edges = [(1,2),(0,1),(0,0),(0,0)]
            for i in [0,1]:
                edge_list = edges[i]
                _mat_RHS[corner_interior_IDs[i]] -= _delta_t*numpy.sum(nine_node_map[_interior_corners[i],:][corners_antimasks[i,edge_list]]*elev[_interior_corners[i]+modulator_mask[corners_antimasks[i,edge_list]]])
            #make dummy array objects for the x,y coords in coo creation of _operating_matrix
            bottom_op_mat_row_add = numpy.empty(0)
            bottom_op_mat_col_add = numpy.empty(0)
            bottom_op_mat_data_add = numpy.empty(0)
        elif self.bottom_flag == 4 or self.bottom_flag == 2: #i.e., fixed zero gradient (4) or more general case...
            bottom_op_mat_row_add = numpy.empty((bottom_interior_IDs.size*3+6)) #was 4
            bottom_op_mat_col_add = numpy.empty((bottom_interior_IDs.size*3+6))
            bottom_op_mat_data_add = numpy.empty((bottom_interior_IDs.size*3+6))
            #Equivalent to fixed gradient, but the gradient is zero, so material only goes in the linked cell(i.e., each cell in the op_mat edges points back to itself).
            bottom_op_mat_row_add[:(bottom_interior_IDs.size*3)] = numpy.repeat(bottom_interior_IDs,3)
            bottom_op_mat_col_add[:(bottom_interior_IDs.size*3)] = self.realIDtointerior(self.operating_matrix_ID_map[self.bottom_interior_IDs,:][:,self.bottom_mask[0:3]]).flatten()
            bottom_op_mat_data_add[:(bottom_interior_IDs.size*3)] = _delta_t*(nine_node_map[_bottom_list,:][:,bottom_antimask]).flatten()
            #...& the corners
            this_corner_coords = numpy.array([0,1])
            #order is bottom 2 lower left, bottom 2 lower right, lower left true corner, lower right true corner.
            bottom_op_mat_row_add[-6:-2] = numpy.repeat(corner_interior_IDs[this_corner_coords],2)
            bottom_op_mat_col_add[-6:-2] = self.operating_matrix_corner_int_IDs[this_corner_coords.reshape(2,1),this_corner_coords].flatten()
            bottom_op_mat_row_add[-2:] =  corner_interior_IDs[this_corner_coords] #new
            bottom_op_mat_col_add[-2:] = self.operating_matrix_corner_int_IDs[(this_corner_coords[0],this_corner_coords[0]),(this_corner_coords[1],this_corner_coords[1])].flatten()
            bottom_op_mat_data_add[-6:-4] = _delta_t*nine_node_map[_interior_corners[0],:][corners_antimasks[0,[1,2]]].flatten()
            bottom_op_mat_data_add[-4:-2] = _delta_t*nine_node_map[_interior_corners[1],:][corners_antimasks[1,[0,1]]].flatten()
            bottom_op_mat_data_add[-2] = _delta_t*nine_node_map[_interior_corners[0],:][corners_antimasks[0,0]]
            bottom_op_mat_data_add[-1] = _delta_t*nine_node_map[_interior_corners[1],:][corners_antimasks[1,2]]
            if self.bottom_flag == 2:
                #Read the offsets from the map we made in the __init__, use them as constant terms, incorporated into RHS
                _mat_RHS[bottom_interior_IDs] -= _delta_t*numpy.sum(nine_node_map[_bottom_list,:][:,bottom_antimask]*self.fixed_gradient_offset_map[_bottom_list.reshape((len(_bottom_list),1))+(modulator_mask[bottom_antimask]).reshape(1,3)], axis=1) #note the broadcasting to (nedge,3) in final fancy index   
                #...& the corners
                edges = [(1,2),(0,1),(0,0),(0,0)]
                for i in [0,1]:
                    edge_list = edges[i]
                    _mat_RHS[corner_interior_IDs[i]] -= _delta_t*numpy.sum(nine_node_map[_interior_corners[i],:][corners_antimasks[i,edge_list]]*self.fixed_gradient_offset_map[_interior_corners[i]+modulator_mask[corners_antimasks[i,edge_list]]])
        elif self.bottom_flag == 3:
            #This will handle both top and bottom BCs...
            bottom_op_mat_row_add = numpy.empty((bottom_interior_IDs.size*3+6)) #was 4
            bottom_op_mat_col_add = numpy.empty((bottom_interior_IDs.size*3+6))
            bottom_op_mat_data_add = numpy.empty((bottom_interior_IDs.size*3+6))
            bottom_op_mat_row_add[:(bottom_interior_IDs.size*3)] = numpy.repeat(bottom_interior_IDs,3) #...put the values in the same places in the operating matrix...
            bottom_op_mat_col_add[:(bottom_interior_IDs.size*3)] = self.realIDtointerior(self.operating_matrix_ID_map[self.top_interior_IDs,:][:,self.top_mask[3:6]]).flatten()
            bottom_op_mat_data_add[:(bottom_interior_IDs.size*3)] = _delta_t*(nine_node_map[_bottom_list,:][:,bottom_antimask]).flatten() #...but the values refer to the TOP of the grid
            top_op_mat_row_add = numpy.empty((top_interior_IDs.size*3+6))
            top_op_mat_col_add = numpy.empty((top_interior_IDs.size*3+6))
            top_op_mat_data_add = numpy.empty((top_interior_IDs.size*3+6))
            top_op_mat_row_add[:(top_interior_IDs.size*3)] = numpy.repeat(top_interior_IDs,3)
            top_op_mat_col_add[:(top_interior_IDs.size*3)] = self.realIDtointerior(self.operating_matrix_ID_map[self.bottom_interior_IDs,:][:,self.bottom_mask[0:3]]).flatten()
            top_op_mat_data_add[:(top_interior_IDs.size*3)] = _delta_t*(nine_node_map[_top_list,:][:,top_antimask]).flatten()
            #& the corners
            bottom_corner_coords = numpy.array([0,1])
            top_corner_coords = numpy.array([2,3])
            bottom_op_mat_row_add[-6:-2] = numpy.repeat(corner_interior_IDs[bottom_corner_coords],2)
            bottom_op_mat_col_add[-6:-2] = self.operating_matrix_corner_int_IDs[top_corner_coords.reshape(2,1),top_corner_coords].flatten()
            bottom_op_mat_row_add[-2:] =  corner_interior_IDs[bottom_corner_coords]
            bottom_op_mat_col_add[-2:] = self.operating_matrix_corner_int_IDs[(top_corner_coords[0],top_corner_coords[0]),(top_corner_coords[1],top_corner_coords[1])].flatten()
            bottom_op_mat_data_add[-6:-4] = _delta_t*nine_node_map[_interior_corners[0],:][corners_antimasks[0,[1,2]]].flatten()
            bottom_op_mat_data_add[-4:-2] = _delta_t*nine_node_map[_interior_corners[1],:][corners_antimasks[1,[0,1]]].flatten()
            bottom_op_mat_data_add[-2] = _delta_t*nine_node_map[_interior_corners[0],:][corners_antimasks[0,0]]
            bottom_op_mat_data_add[-1] = _delta_t*nine_node_map[_interior_corners[1],:][corners_antimasks[1,2]]
            top_op_mat_row_add[-6:-2] = numpy.repeat(corner_interior_IDs[top_corner_coords],2)
            top_op_mat_col_add[-6:-2] = self.operating_matrix_corner_int_IDs[bottom_corner_coords.reshape(2,1),bottom_corner_coords].flatten()
            top_op_mat_row_add[-2:] =  corner_interior_IDs[top_corner_coords] #new
            top_op_mat_col_add[-2:] = self.operating_matrix_corner_int_IDs[(bottom_corner_coords[0],bottom_corner_coords[0]),(bottom_corner_coords[1],bottom_corner_coords[1])].flatten()
            top_op_mat_data_add[-6:-4] = _delta_t*nine_node_map[_interior_corners[2],:][corners_antimasks[2,[3,4]]].flatten()
            top_op_mat_data_add[-4:-2] = _delta_t*nine_node_map[_interior_corners[3],:][corners_antimasks[3,[2,3]]].flatten()
            top_op_mat_data_add[-2] = _delta_t*nine_node_map[_interior_corners[2],:][corners_antimasks[2,2]]
            top_op_mat_data_add[-1] = _delta_t*nine_node_map[_interior_corners[3],:][corners_antimasks[3,4]]
        else:
            raise NameError('Something is very wrong with your boundary conditions...!')

        if self.top_flag == 1:
            #goes to RHS only
            _mat_RHS[top_interior_IDs] -= _delta_t*numpy.sum(nine_node_map[_top_list,:][:,top_antimask]*elev[_top_list.reshape((len(_top_list),1))+(modulator_mask[top_antimask]).reshape(1,3)], axis=1)
            #...& the corners
            edges = [(0,0),(0,0),(3,4),(2,3)]
            for i in [2,3]:
                edge_list = edges[i]
                _mat_RHS[corner_interior_IDs[i]] -= _delta_t*numpy.sum(nine_node_map[_interior_corners[i],:][corners_antimasks[i,edge_list]]*elev[_interior_corners[i]+modulator_mask[corners_antimasks[i,edge_list]]])
            top_op_mat_row_add = numpy.empty(0)
            top_op_mat_col_add = numpy.empty(0)
            top_op_mat_data_add = numpy.empty(0)
        elif self.top_flag == 4 or self.top_flag == 2:
            top_op_mat_row_add = numpy.empty((top_interior_IDs.size*3+6))
            top_op_mat_col_add = numpy.empty((top_interior_IDs.size*3+6))
            top_op_mat_data_add = numpy.empty((top_interior_IDs.size*3+6))
            #Equivalent to fixed gradient, but the gradient is zero, so material only goes in the linked cell(i.e., each cell in the op_mat edges points back to itself).
            top_op_mat_row_add[:(top_interior_IDs.size*3)] = numpy.repeat(top_interior_IDs,3)
            top_op_mat_col_add[:(top_interior_IDs.size*3)] = self.realIDtointerior(self.operating_matrix_ID_map[self.top_interior_IDs,:][:,self.top_mask[3:6]]).flatten()
            top_op_mat_data_add[:(top_interior_IDs.size*3)] = _delta_t*(nine_node_map[_top_list,:][:,top_antimask]).flatten()
            #...& the corners
            this_corner_coords = numpy.array([2,3])
            top_op_mat_row_add[-6:-2] = numpy.repeat(corner_interior_IDs[this_corner_coords],2)
            top_op_mat_col_add[-6:-2] = self.operating_matrix_corner_int_IDs[this_corner_coords.reshape(2,1),this_corner_coords].flatten()
            top_op_mat_row_add[-2:] =  corner_interior_IDs[this_corner_coords] #new
            top_op_mat_col_add[-2:] = self.operating_matrix_corner_int_IDs[(this_corner_coords[0],this_corner_coords[0]),(this_corner_coords[1],this_corner_coords[1])].flatten()
            top_op_mat_data_add[-6:-4] = _delta_t*nine_node_map[_interior_corners[2],:][corners_antimasks[2,[3,4]]].flatten()
            top_op_mat_data_add[-4:-2] = _delta_t*nine_node_map[_interior_corners[3],:][corners_antimasks[3,[2,3]]].flatten()
            top_op_mat_data_add[-2] = _delta_t*nine_node_map[_interior_corners[2],:][corners_antimasks[2,2]]
            top_op_mat_data_add[-1] = _delta_t*nine_node_map[_interior_corners[3],:][corners_antimasks[3,4]]
            if self.top_flag == 2:
                _mat_RHS[top_interior_IDs] -= _delta_t*numpy.sum(nine_node_map[_top_list,:][:,top_antimask]*self.fixed_gradient_offset_map[_top_list.reshape((len(_top_list),1))+(modulator_mask[top_antimask]).reshape(1,3)], axis=1)
                #...& the corners
                edges = [(0,0),(0,0),(3,4),(2,3)]
                for i in [2,3]:
                    edge_list = edges[i]
                    _mat_RHS[corner_interior_IDs[i]] -= _delta_t*numpy.sum(nine_node_map[_interior_corners[i],:][corners_antimasks[i,edge_list]]*self.fixed_gradient_offset_map[_interior_corners[i]+modulator_mask[corners_antimasks[i,edge_list]]])
        elif self.top_flag == 3:
            pass #dealt with above
        else:
            raise NameError('Something is very wrong with your boundary conditions...!')

        if self.left_flag == 1:
            #goes to RHS only
            _mat_RHS[left_interior_IDs] -= _delta_t*numpy.sum(nine_node_map[_left_list,:][:,left_antimask]*elev[_left_list.reshape((len(_left_list),1))+(modulator_mask[left_antimask]).reshape(1,3)], axis=1)
            #...& the corners
            edges = [(3,4),(0,0),(0,1),(0,0)]
            for i in [0,2]:
                edge_list = edges[i]
                _mat_RHS[corner_interior_IDs[i]] -= _delta_t*numpy.sum(nine_node_map[_interior_corners[i],:][corners_antimasks[i,edge_list]]*elev[_interior_corners[i]+modulator_mask[corners_antimasks[i,edge_list]]])
            left_op_mat_row_add = numpy.empty(0)
            left_op_mat_col_add = numpy.empty(0)
            left_op_mat_data_add = numpy.empty(0)
        elif self.left_flag == 4 or self.left_flag == 2:
            left_op_mat_row_add = numpy.empty((left_interior_IDs.size*3+4))
            left_op_mat_col_add = numpy.empty((left_interior_IDs.size*3+4))
            left_op_mat_data_add = numpy.empty((left_interior_IDs.size*3+4))
            #Equivalent to fixed gradient, but the gradient is zero, so material only goes in the linked cell(i.e., each cell in the op_mat edges points back to itself).
            left_op_mat_row_add[:(left_interior_IDs.size*3)] = numpy.repeat(left_interior_IDs,3)
            left_op_mat_col_add[:(left_interior_IDs.size*3)] = self.realIDtointerior(self.operating_matrix_ID_map[self.left_interior_IDs,:][:,self.left_mask[::2]]).flatten()
            left_op_mat_data_add[:(left_interior_IDs.size*3)] = _delta_t*(nine_node_map[_left_list,:][:,left_antimask]).flatten()
            #...& the corners
            this_corner_coords = numpy.array([0,2])
            left_op_mat_row_add[-4:] = numpy.repeat(corner_interior_IDs[this_corner_coords],2)
            left_op_mat_col_add[-4:] = self.operating_matrix_corner_int_IDs[this_corner_coords.reshape(2,1),this_corner_coords].flatten()
            left_op_mat_data_add[-4:-2] = _delta_t*nine_node_map[_interior_corners[0],:][corners_antimasks[0,[3,4]]].flatten()
            left_op_mat_data_add[-2:] = _delta_t*nine_node_map[_interior_corners[2],:][corners_antimasks[2,[0,1]]].flatten()
            if self.left_flag == 2:
                _mat_RHS[left_interior_IDs] -= _delta_t*numpy.sum(nine_node_map[_left_list,:][:,left_antimask]*self.fixed_gradient_offset_map[_left_list.reshape((len(_left_list),1))+(modulator_mask[left_antimask]).reshape(1,3)], axis=1)
                #...& the corners
                edges = [(3,4),(0,0),(0,1),(0,0)]
                for i in [0,2]:
                    edge_list = edges[i]
                    _mat_RHS[corner_interior_IDs[i]] -= _delta_t*numpy.sum(nine_node_map[_interior_corners[i],:][corners_antimasks[i,edge_list]]*self.fixed_gradient_offset_map[_interior_corners[i]+modulator_mask[corners_antimasks[i,edge_list]]])
        elif self.left_flag ==3:
            left_op_mat_row_add = numpy.empty((left_interior_IDs.size*3+4))
            left_op_mat_col_add = numpy.empty((left_interior_IDs.size*3+4))
            left_op_mat_data_add = numpy.empty((left_interior_IDs.size*3+4))
            left_op_mat_row_add[:(left_interior_IDs.size*3)] = numpy.repeat(left_interior_IDs,3)
            left_op_mat_col_add[:(left_interior_IDs.size*3)] = self.realIDtointerior(self.operating_matrix_ID_map[self.right_interior_IDs,:][:,self.right_mask[1::2]]).flatten()
            left_op_mat_data_add[:(left_interior_IDs.size*3)] = _delta_t*(nine_node_map[_left_list,:][:,left_antimask]).flatten()
            right_op_mat_row_add = numpy.empty((right_interior_IDs.size*3+4))
            right_op_mat_col_add = numpy.empty((right_interior_IDs.size*3+4))
            right_op_mat_data_add = numpy.empty((right_interior_IDs.size*3+4))
            right_op_mat_row_add[:(right_interior_IDs.size*3)] = numpy.repeat(right_interior_IDs,3)
            right_op_mat_col_add[:(right_interior_IDs.size*3)] = self.realIDtointerior(self.operating_matrix_ID_map[self.left_interior_IDs,:][:,self.left_mask[::2]]).flatten()
            right_op_mat_data_add[:(right_interior_IDs.size*3)] = _delta_t*(nine_node_map[_right_list,:][:,right_antimask]).flatten()
            #& the corners
            left_corner_coords = numpy.array([0,2])
            right_corner_coords = numpy.array([1,3])
            left_op_mat_row_add[-4:] = numpy.repeat(corner_interior_IDs[left_corner_coords],2)
            left_op_mat_col_add[-4:] = self.operating_matrix_corner_int_IDs[right_corner_coords.reshape(2,1),right_corner_coords].flatten()
            left_op_mat_data_add[-4:-2] = _delta_t*nine_node_map[_interior_corners[0],:][corners_antimasks[0,[3,4]]].flatten()
            left_op_mat_data_add[-2:] = _delta_t*nine_node_map[_interior_corners[2],:][corners_antimasks[2,[0,1]]].flatten()
            right_op_mat_row_add[-4:] = numpy.repeat(corner_interior_IDs[right_corner_coords],2)
            right_op_mat_col_add[-4:] = self.operating_matrix_corner_int_IDs[left_corner_coords.reshape(2,1),left_corner_coords].flatten()
            right_op_mat_data_add[-4:-2] = _delta_t*nine_node_map[_interior_corners[1],:][corners_antimasks[1,[3,4]]].flatten()
            right_op_mat_data_add[-2:] = _delta_t*nine_node_map[_interior_corners[3],:][corners_antimasks[3,[0,1]]].flatten()
        else:
            raise NameError('Something is very wrong with your boundary conditions...!')

        if self.right_flag == 1:
            #goes to RHS only
            _mat_RHS[right_interior_IDs] -= _delta_t*numpy.sum(nine_node_map[_right_list,:][:,right_antimask]*elev[_right_list.reshape((len(_right_list),1))+(modulator_mask[right_antimask]).reshape(1,3)], axis=1)
            #...& the corners
            edges = [(0,0),(3,4),(0,0),(0,1)]
            for i in [1,3]:
                edge_list = edges[i]
                _mat_RHS[corner_interior_IDs[i]] -= _delta_t*numpy.sum(nine_node_map[_interior_corners[i],:][corners_antimasks[i,edge_list]]*elev[_interior_corners[i]+modulator_mask[corners_antimasks[i,edge_list]]])
            right_op_mat_row_add = numpy.empty(0)
            right_op_mat_col_add = numpy.empty(0)
            right_op_mat_data_add = numpy.empty(0)
        elif self.right_flag == 4 or self.right_flag == 2:
            right_op_mat_row_add = numpy.empty((right_interior_IDs.size*3+4))
            right_op_mat_col_add = numpy.empty((right_interior_IDs.size*3+4))
            right_op_mat_data_add = numpy.empty((right_interior_IDs.size*3+4))
            #Equivalent to fixed gradient, but the gradient is zero, so material only goes in the linked cell(i.e., each cell in the op_mat edges points back to itself).
            right_op_mat_row_add[:(right_interior_IDs.size*3)] = numpy.repeat(right_interior_IDs,3)
            right_op_mat_col_add[:(right_interior_IDs.size*3)] = self.realIDtointerior(self.operating_matrix_ID_map[self.right_interior_IDs,:][:,self.right_mask[1::2]]).flatten()
            right_op_mat_data_add[:(right_interior_IDs.size*3)] = _delta_t*(nine_node_map[_right_list,:][:,right_antimask]).flatten()
            #...& the corners
            this_corner_coords = numpy.array([1,3])
            right_op_mat_row_add[-4:] = numpy.repeat(corner_interior_IDs[this_corner_coords],2)
            right_op_mat_col_add[-4:] = self.operating_matrix_corner_int_IDs[this_corner_coords.reshape(2,1),this_corner_coords].flatten()
            right_op_mat_data_add[-4:-2] = _delta_t*nine_node_map[_interior_corners[1],:][corners_antimasks[1,[3,4]]].flatten()
            right_op_mat_data_add[-2:] = _delta_t*nine_node_map[_interior_corners[3],:][corners_antimasks[3,[0,1]]].flatten()
            if self.right_flag == 2:
                _mat_RHS[right_interior_IDs] -= _delta_t*numpy.sum(nine_node_map[_right_list,:][:,right_antimask]*self.fixed_gradient_offset_map[_right_list.reshape((len(_right_list),1))+(modulator_mask[right_antimask]).reshape(1,3)], axis=1)
                #...& the corners
                edges = [(0,0),(3,4),(0,0),(0,1)]
                for i in [1,3]:
                    edge_list = edges[i]
                    _mat_RHS[corner_interior_IDs[i]] -= _delta_t*numpy.sum(nine_node_map[_interior_corners[i],:][corners_antimasks[i,edge_list]]*self.fixed_gradient_offset_map[_interior_corners[i]+modulator_mask[corners_antimasks[i,edge_list]]])
        elif self.top_flag == 3:
            pass #dealt with above
        else:
            raise NameError('Something is very wrong with your boundary conditions...!')

        ###new approach using COO sparse matrix requires we build the matrix only now...
        self._operating_matrix = sparse.coo_matrix((numpy.concatenate((core_op_mat_data,corners_op_mat_data,bottom_op_mat_data,top_op_mat_data,left_op_mat_data,right_op_mat_data,bottom_op_mat_data_add,top_op_mat_data_add,left_op_mat_data_add,right_op_mat_data_add)),
            (numpy.concatenate((core_op_mat_row,corners_op_mat_row,bottom_op_mat_row,top_op_mat_row,left_op_mat_row,right_op_mat_row,bottom_op_mat_row_add,top_op_mat_row_add,left_op_mat_row_add,right_op_mat_row_add)),
            numpy.concatenate((core_op_mat_col,corners_op_mat_col,bottom_op_mat_col,top_op_mat_col,left_op_mat_col,right_op_mat_col,bottom_op_mat_col_add,top_op_mat_col_add,left_op_mat_col_add,right_op_mat_col_add)))),
            shape=(n_interior_nodes, n_interior_nodes)).tocsr()
        self._mat_RHS = _mat_RHS


#These methods translate ID numbers between arrays of differing sizes                
    def realIDtointerior(self, ID):
        ncols = self.ncols
        interior_ID = (ID//ncols - 1)*(ncols-2) + (ID%ncols) - 1
        if numpy.any(interior_ID < 0) or numpy.any(interior_ID >= self.ninteriornodes):
            print "One of the supplied nodes was outside the interior grid!"
            raise NameError()
        else:
            return interior_ID.astype(int)
        
    def interiorIDtoreal(self, ID):
        IGW = self.interior_grid_width
        real_ID = (ID//IGW + 1) * self.ncols + (ID%IGW) + 1
        assert numpy.all(real_ID < self.nnodes)
        return real_ID.astype(int)
    
    def realIDtocore(self, ID):
        ncols = self.ncols
        core_ID = (ID//ncols - 2)*(ncols-4) + (ID%ncols) - 2
        if numpy.any(core_ID < 0) or numpy.any(core_ID >= self.ncorenodes):
            print "One of the supplied nodes was outside the core grid!"
            raise NameError()
        else:
            return core_ID.astype(int)

    def coreIDtoreal(self, ID):
        CCW = self.core_cell_width
        real_ID = (ID//CCW + 2) * self.ncols + (ID%CCW) + 2
        assert numpy.all(real_ID < self.nnodes)
        return real_ID.astype(int)

    def interiorIDtocore(self, ID):
        IGW = self.interior_grid_width
        core_ID = (ID//IGW - 1)*(self.ncols-4) + (ID%IGW) - 1
        if numpy.any(core_ID < 0) or numpy.any(core_ID >= self.ncorenodes):
            print "One of the supplied nodes was outside the core grid!"
            raise NameError()
        else:
            return core_ID.astype(int)
        
    def coreIDtointerior(self, ID):
        CCW = self.core_cell_width
        interior_ID = (ID//CCW + 1) * (self.ncols-2) + (ID%CCW) + 1
        assert numpy.all(interior_ID < self.ninteriornodes)
        return interior_ID.astype(int)

    def diffuse(self, grid_in, elapsed_time, num_uplift_implicit_comps = 1):
        """
        This is the primary method of the class. Call it to perform an iteration
        of the model. Takes *grid_in*, the model grid, and *elapsed_time*, the
        total model time elapsed so far.
        
        *grid_in* must contain the field to diffuse, which defaults to
        'topographic_elevation'. This can be overridden with the 
        values_to_diffuse property in the input file.
        
        See the class docstring for a list of the other properties necessary
        in the input file for this component to run.
        
        Note that the implicit nature of this component requires it to
        incorporate uplift into its execution in order to stay stable.
        If you only have one module that requires this, do not add uplift
        manually in your loop; this method will include uplift automatically.
        
        If more than one of your components has this requirement, set
        *num_uplift_implicit_comps* to the total number of components that
        do.
        """
        if self.internal_uplifts:
            #this is adhoc to fix for the duration of Germany visit
            self._uplift = self.inputs.read_float('uplift_rate')
            self._delta_t = self.timestep_in
            self.set_variables(self.grid)
            _interior_elevs = linalg.spsolve(self._operating_matrix, self._mat_RHS)
            self.grid['node'][self.values_to_diffuse][self.interior_IDs_as_real] = _interior_elevs
            #self.grid['node'][self.values_to_diffuse][grid_in.get_boundary_nodes()] += self._uplift
            grid_in=self.grid
        else:
            self.gear_timestep(self.timestep_in, grid_in)
            for i in xrange(self.internal_repeats):
                grid_in['node'][self.values_to_diffuse] = self.grid['node'][self.values_to_diffuse] + self.uplift_per_step
            #Initialize the variables for the step:
                self.set_variables(grid_in)
                #Solve interior of grid:
                _interior_elevs = linalg.spsolve(self._operating_matrix, self._mat_RHS)
                #print 'solved the matrix'
                #print _interior_elevs.shape
                #this fn solves Ax=B for x
                
                #Handle the BC cells; test common cases first for speed
                self.grid['node'][self.values_to_diffuse][self.interior_IDs_as_real] = _interior_elevs
                
                #if BC==1 or BC==4, don't need to take any action; in both cases the values are unchanged.
                if self.fixed_grad_BCs_present:
                    self.grid['node'][self.values_to_diffuse][
                        grid_in.fixed_gradient_node_properties['boundary_node_IDs']] = self.grid[
                        'node'][self.values_to_diffuse][self.grid.fixed_gradient_node_properties[
                        'anchor_node_IDs']] + self.grid.fixed_gradient_node_properties['values_to_add']
                if self.looped_BCs_present:
                    self.grid['node'][self.values_to_diffuse][
                        self.grid.looped_node_properties['boundary_node_IDs']] = self.grid[
                        'node'][self.values_to_diffuse][self.grid.looped_node_properties[
                        'linked_node_IDs']]
                    #raise ValueError #need to update this once workflow for looped BCs is established
        
        return self.grid



from landlab.components.flow_routing.route_flow_dn import FlowRouter
from landlab.components.stream_power.fastscape_stream_power import SPEroder
from landlab import ModelParameterDictionary
from landlab.plot import channel_profile as prf
from landlab.plot.imshow import imshow_node_grid

from landlab import RasterModelGrid
import numpy as np
import pylab

from time import time

#get the needed properties to build the grid:
input_file = './stream_power_params.txt'
inputs = ModelParameterDictionary(input_file)
nrows = inputs.read_int('nrows')
ncols = inputs.read_int('ncols')
dx = inputs.read_float('dx')
leftmost_elev = inputs.read_float('leftmost_elevation')
initial_slope = inputs.read_float('initial_slope')
uplift_rate = inputs.read_float('uplift_rate')
runtime = inputs.read_float('total_time')
dt = inputs.read_float('dt')
nt = int(runtime//dt)
uplift_per_step = uplift_rate * dt

#instantiate the grid object
mg = RasterModelGrid(nrows, ncols, dx)

##create the elevation field in the grid:
#create the field
mg.create_node_array_zeros('topographic_elevation')
z = mg.create_node_array_zeros() + leftmost_elev
z += initial_slope*np.amax(mg.node_y) - initial_slope*mg.node_y
#put these values plus roughness into that field
mg['node'][ 'topographic_elevation'] = z + np.random.rand(len(z))/100000.

#set up its boundary conditions (bottom, left, top, right is inactive)
mg.set_closed_boundaries_at_grid_edges(False, True, False, True)

# Display a message
print 'Running ...' 

#instantiate the components:
fr = FlowRouter(mg)
sp = SPEroder(mg, input_file)

time_on = time()
#perform the loops:
for i in xrange(nt):
    mg['node']['topographic_elevation'][mg.core_nodes] += uplift_per_step
    mg = fr.route_flow(grid=mg)
    mg = sp.erode(mg)

    #plot long profiles along channels
    pylab.figure(6)
    profile_IDs = prf.channel_nodes(mg, mg.at_node['steepest_slope'],
            mg.at_node['drainage_area'], mg.at_node['upstream_ID_order'],
            mg.at_node['flow_receiver'])
    dists_upstr = prf.get_distances_upstream(mg, len(mg.at_node['steepest_slope']),
            profile_IDs, mg.at_node['links_to_flow_receiver'])
    prf.plot_profiles(dists_upstr, profile_IDs, mg.at_node['topographic_elevation'])
    print 'Completed loop ', i
 
print 'Completed the simulation. Plotting...'

time_off = time()

#Finalize and plot
# Clear previous plots
pylab.figure(1)
pylab.close()
pylab.figure(1)
im = imshow_node_grid(mg, 'water_discharges', cmap='Blues')  # display a colored image

pylab.figure(2)
im = imshow_node_grid(mg, 'topographic_elevation')  # display a colored image

elev = mg['node']['topographic_elevation']
elev_r = mg.node_vector_to_raster(elev)
pylab.figure(3)
im = pylab.plot(mg.dx*np.arange(nrows), elev_r[:,int(ncols//2)])
pylab.title('N-S cross_section')

pylab.figure(4)
im = pylab.plot(mg.dx*np.arange(ncols), elev_r[int(nrows//4),:])
pylab.title('E-W cross_section')

drainage_areas = mg['node']['drainage_area'][mg.get_interior_nodes()]
steepest_slopes = mg['node']['steepest_slope'][mg.get_interior_nodes()]
pylab.figure(5)
pylab.loglog(drainage_areas, steepest_slopes, 'x')
pylab.xlabel('Upstream drainage area, m^2')
pylab.ylabel('Maximum slope')

print('Done.')
print 'Time: ', time_off-time_on

pylab.show()



""" generate_fire.py

This component generates random numbers using the Weibull distribution
(Weibull, 1951). No particular units must be used, but it was written with
the fire recurrence units in time (yrs).

Using the Weibull Distribution assumes two things: All elements within the study
area have the same fire regime. Each element must have (on average) a constant
fire regime during the time span of the study. 

As of Sept. 2013, fires are considered instantaneous events independent of
other fire events in the time series. 

Written by Jordan Marie Adams, 2013.

"""

import os
from random import weibullvariate
from scipy import special
from landlab import ModelParameterDictionary

_DEFAULT_INPUT_FILE = os.path.join(os.path.dirname(__file__), 'fire.txt')

class FireGenerator:

    def __init__(self, input_file=None):
        """All initial values are set to zero until initalized
        using the initialize() method which reads in data using
        the ModelParameterDictionary and sets the random variables
        according to their respective distribution.
        
        REQUIRED PARAMETERS
        ------------------
        
        Shape Parameter: Describes the skew of the Weibull distribution. 
        If shape < 3.5, data skews left.
        If shape == 3.5, data is normal.
        If shape > 3.5, data skews right.
        
        Scale Parameter: Describes the peak of the Weibull distribution, 
        located at 63.5% value of the cumulative distribution function. If unknown,
        it can be found using mean fire recurrence value and the get_scale_parameter()
        method described later.
        
        Mean Fire Recurrence : Average recurrence for a given area, elevation, veg type, etc.
        
        Total Run Time : Total model run time
        
        Delta T : Model time step.
        
        
        -------
        Time to Next Fire: Value generated from the random.weibullvariate() function based
        on the scale and shape parameters
        """        
        # We import methods from ModelParameterDictionary
        # to read the parameters from the input file.
        
        #If the scale parameter is unknown, it can be found using the mean
        # fire recurrence value, which MUST be known or estimated to run the
        # get_scale_parameter() method."""
        
        MPD = ModelParameterDictionary()

        if input_file is None:
            input_file = _DEFAULT_INPUT_FILE
        MPD.read_from_file(input_file)
        
        self.shape_parameter = MPD.read_float("SHAPE_PARAMETER")
        self.scale_parameter = MPD.read_float("SCALE_PARAMETER")
        self.mean_fire_recurrence = MPD.read_float("MEAN_FIRE_RECURRENCE")
        self.total_run_time = MPD.read_float("RUN_TIME")
        self.delta_t = MPD.read_int("DELTA_T")
        self.time_to_next_fire = 0.0

    def get_scale_parameter(self):
        
        """ If the scale factor is unknown, we can draw it from the mean
        fire recurrence interval, given the following equation:
            
        This ONLY works if we have the shape parameter. Generally, the shape
        parameter will be greater than 3.5, creating a distribution that is skewed
        to the higher values (skewed to the right). 
            
        Mean_fire_recurrence = scale_parameter*(gamma_function(1+(1/shape)))
        
        :returns: scale_parameter as a float"""
        
        if self.scale_parameter == 0.0:   
            shape_in_gamma_func = float(1+(1/self.shape_parameter))
            gamma_func = special.gamma(shape_in_gamma_func)
            self.scale_parameter = (self.mean_fire_recurrence/gamma_func)
            return self.scale_parameter
        else:
            return self.scale_parameter
            
    def generate_fire_recurrence(self):
        
        """ Finds the time to next fire (fire recurrence) based on the scale parameter (63.5% of
        fire Weibull distribution) and the shape parameter (describes the skew of the histogram, shape = 3.5
        represents a normal distribution).
        
        Rounds the time to next fire to 4 significant figures, for neatness.
        
        :returns: time_to_next_fire as a float"""
        
        self.time_to_next_fire = round(weibullvariate(self.scale_parameter, self.shape_parameter),2)
        return self.time_to_next_fire
        
    def generate_fire_time_series(self):
        
        """Allows for a series of fire events to be generated given 
        a total time. 
        
        Created for situations where total run time is definite, and number
        of fires can change across different runs. 
        
        :creates: array with several fire events, all values are float"""
        
        self.fire_events =[]
        event = self.generate_fire_recurrence()
        end_event = event + 365.0
        self.fire_events.append([event, end_event])
        t = 0
        i = 0
        while t <= self.total_run_time:
            fire = self.generate_fire_recurrence()
            start_fire = self.fire_events[i][0] + (fire)
            end_fire = start_fire + (365.0)       
            self.fire_events.append([start_fire, end_fire])
            t += end_fire
            i+=1
       
    def update(self):        
        """Update function allows us to update the value of "time_to_next_fire" by
        re-calling the generate_fire_reccurence() function.
        
        Created for instances when a definite number of fires need to
        be generated.
        
        :returns: updated value for time_to_next_fire as a float"""
        
        self.time_to_next_fire = self.generate_fire_recurrence()
        return self.time_to_next_fire



import os
import sys
from numpy.testing import Tester


def show_system_info():
    import landlab
    import nose

    print('landlab version %s' % landlab.__version__)
    landlab_dir = os.path.dirname(landlab.__file__)
    print('landlab is installed in %s' % landlab_dir)

    print('Python version %s' % sys.version.replace('\n', ''))
    print('nose version %d.%d.%d' % nose.__versioninfo__)


class LandlabTester(Tester):
    excludes = ['examples']

    def __init__(self, package=None, raise_warnings='develop'):
        package_name = None
        if package is None:
            f = sys._getframe(1)
            package_path = f.f_locals.get('__file__', None)
            if package_path is None:
                raise AssertionError
            package_path = os.path.dirname(package_path)
            package_name = f.f_locals.get('__name__', None)
        elif isinstance(package, type(os)):
            package_path = os.path.dirname(package.__file__)
            package_name = getattr(package, '__name__', None)
        else:
            package_path = str(package)

        self.package_path = os.path.abspath(package_path)
        
        # Find the package name under test; this name is used to limit coverage
        # reporting (if enabled).
        if package_name is None:
            #package_name = get_package_name(package_path)
            package_name = 'landlab'
        self.package_name = package_name
        
        # Set to "release" in constructor in maintenance branches.
        self.raise_warnings = raise_warnings


    def test(self, **kwds):
        kwds.setdefault('verbose', 2)
        kwds.setdefault('doctests', 1)
        kwds.setdefault('coverage', 1)
        kwds.setdefault('extra_argv', ['-x', 'where=landlab'])
        show_system_info()
        return super(LandlabTester, self).test(**kwds)



#! /usr/env/python
"""
oriented_hex_lca.py: simple hexagonal Landlab cellular automaton

This file defines the OrientedHexLCA class, which is a sub-class of 
LandlabCellularAutomaton that implements a simple, non-oriented, hex-grid
CA. Like its parent class, OrientedHexLCA implements a continuous-time, stochastic,
pair-based CA. The hex grid has 3 principal directions, rather than 2 for a
raster. Hex grids are often used in CA models because of their symmetry.

Created GT Sep 2014
"""

from numpy import zeros
from landlab_ca import LandlabCellularAutomaton, Transition
import landlab


class OrientedHexLCA(LandlabCellularAutomaton):
    
    def __init__(self, model_grid, node_state_dict, transition_list,
                 initial_node_states, prop_data=None, prop_reset_value=None):
        
        # Make sure caller has sent the right grid type        
        assert (type(model_grid) is landlab.grid.hex.HexModelGrid), \
               'model_grid must be a Landlab HexModelGrid'
               
        # Somehow test to make sure the grid links have been re-oriented to
        # point up/right (-45 to +135 degrees clockwise relative to vertical).
        # Such orientation is ensured when the argument reorient_grid=True is
        # passed to the hex grid constructor.
               
        # Define the number of distinct cell-pair orientations: here 3,
        # representing
        self.number_of_orientations = 3
        
        # Call the LandlabCellularAutomaton.__init__() method to do the rest of
        # the initialization
        super(OrientedHexLCA, self).__init__(model_grid, node_state_dict, 
            transition_list, initial_node_states, prop_data, prop_reset_value)
            

    def setup_array_of_orientation_codes(self):
        """
        Creates and configures an array that contain the orientation code for 
        each active link (and corresponding cell pair).
        
        Parameters
        ----------
        (none)
        
        Returns
        -------
        (none)
        
        Creates
        -------
        self.active_link_orientation : 1D numpy array
        
        Notes
        -----
        This overrides the method of the same name in landlab_ca.py. If the hex
        grid is oriented such that one of the 3 axes is vertical (a 'vertical'
        grid), then the three orientations are:
            0 = vertical (0 degrees clockwise from vertical)
            1 = right and up (60 degrees clockwise from vertical)
            2 = right and down (120 degrees clockwise from vertical)
        If the grid is oriented with one principal axis horizontal ('horizontal'
        grid), then the orientations are:
            0 = up and left (30 degrees counter-clockwise from vertical)
            1 = up and right (30 degrees clockwise from vertical)
            2 = horizontal (90 degrees clockwise from vertical)
        """
        self.active_link_orientation = zeros(self.grid.number_of_active_links, dtype=int)
        for j in range(self.grid.number_of_active_links):
            i = self.grid.active_links[j]
            dy = self.grid.node_y[self.grid.link_tonode[i]]-self.grid.node_y[self.grid.link_fromnode[i]]
            dx = self.grid.node_x[self.grid.link_tonode[i]]-self.grid.node_x[self.grid.link_fromnode[i]]
            if dx <= 0.:
                self.active_link_orientation[j] = 0
            elif dy<=0.:
                self.active_link_orientation[j] = 2
            elif dx>0. and dy>0.:
                self.active_link_orientation[j] = 1
            else:
                assert (False), 'Non-handled link orientation case'


if __name__=='__main__':
    print 'main here'
    from landlab import HexModelGrid
    mg = HexModelGrid(2, 3, 1.0, orientation='vertical', reorient_links=True)
    print mg.number_of_active_links
    nsd = {0 : 'yes', 1 : 'no'}
    xnlist = []
    xnlist.append( Transition( (0,1,0), (1,0,0), 1.0, 'falling' ) )
    nsg = mg.add_zeros('node', 'node_state_grid')
    ohlca = OrientedHexLCA(mg, nsd, xnlist, nsg)
    for i in range(mg.number_of_active_links):
        j = mg.active_links[i]
        print i,j,'(',mg.node_x[mg.link_fromnode[j]],',',mg.node_y[mg.link_fromnode[j]],\
            ')','(',mg.node_x[mg.link_tonode[j]],',',mg.node_y[mg.link_tonode[j]],\
            ')',ohlca.active_link_orientation[i]

